_global_obj_arg_dict = {}
_global_name_dict = {}

assert id(_global_obj_arg_dict) == id(globals()['_global_obj_arg_dict']) == id(locals()['_global_obj_arg_dict'])
assert id(_global_name_dict) == id(globals()['_global_name_dict']) == id(locals()['_global_name_dict'])


def reset_all():
    globals()['_global_obj_arg_dict'] = {}
    globals()['_global_name_dict'] = {}


def reset(key: str):
    globals()[key] = {}


def get_all() -> dict:
    return dict(
        _global_obj_arg_dict=globals()['_global_obj_arg_dict'],
        _global_name_dict=globals()['_global_name_dict']
    )




import tensorflow as tf
import tensorflow_probability as tfp
import unittest
from baconian.core.core import EnvSpec
import numpy as np
from baconian.envs.gym_env import make
from baconian.common.special import *
from baconian.test.tests.set_up.setup import TestTensorflowSetup
import baconian.algo.distribution.mvn as mvn


def describe_sample_tensor_shape(sample_shape, distribution):
    print('Sample shape:', sample_shape)
    print('Returned sample tensor shape:',
          distribution.sample(sample_shape).shape)


def describe_sample_tensor_shapes(distributions, sample_shapes):
    started = False
    for distribution in distributions:
        print(distribution)
        for sample_shape in sample_shapes:
            describe_sample_tensor_shape(sample_shape, distribution)
        print()


def kl_entropy_logprob_from_pat_cody(old_mean, old_var, mean, var, sess, action_dim, action_ph, feed_dict):
    # logvar = tf.reduce_sum(tf.log(var))
    # old_log_var = tf.reduce_sum(tf.log(old_var))
    """
    KL(old|new)
    :param old_mean:
    :param old_var:
    :param mean:
    :param var:
    :param sess:
    :param action_dim:
    :param action_ph:
    :param feed_dict:
    :return:
    """
    logvar = tf.log(var)
    old_log_var = tf.log(old_var)

    log_det_cov_old = tf.reduce_sum(old_log_var)
    log_det_cov_new = tf.reduce_sum(logvar)

    tr_old_new = tf.reduce_sum(tf.exp(old_log_var - logvar))

    kl = 0.5 * tf.reduce_mean(log_det_cov_new - log_det_cov_old + tr_old_new +
                              tf.reduce_sum(tf.square(mean - old_mean) /
                                            tf.exp(logvar), axis=1) -
                              action_dim)

    # kl = 0.5 * (log_det_cov_new - log_det_cov_old + tr_old_new +
    #             tf.reduce_sum(tf.square(mean - old_mean) /
    #                           tf.exp(logvar), axis=0) -
    #             action_dim)

    entropy = 0.5 * (action_dim * (np.log(2 * np.pi) + 1) +
                     tf.reduce_sum(logvar))

    logp = -0.5 * tf.reduce_sum(tf.log(tf.exp(logvar) * 2 * np.pi))
    logp += -0.5 * tf.reduce_sum(tf.square(action_ph - mean) /
                                 tf.exp(logvar), axis=1)
    # logp += -0.5 * np.log(2 * np.pi * action_dim)

    logp_old = -0.5 * tf.reduce_sum(tf.log(tf.exp(old_log_var) * 2 * np.pi))
    logp_old += -0.5 * tf.reduce_sum(tf.square(action_ph - old_mean) /
                                     tf.exp(old_log_var), axis=1)
    # logp_old += -0.5 * np.log(2 * np.pi * action_dim)

    return sess.run([kl, entropy, logp, logp_old], feed_dict=feed_dict)


def kl_entropy_logprob_from_mvn(old_mean, old_var, mean, var, sess, action_dim, action_ph, feed_dict):
    kl = mvn.kl(old_mean, old_var, mean, var, action_dim)
    entropy = mvn.entropy(mean, var, action_dim)
    logp = mvn.log_prob(action_ph, mean, var)
    logp_old = mvn.log_prob(action_ph, old_mean, old_var)
    return sess.run([kl, entropy, logp, logp_old], feed_dict=feed_dict)


class TestTFP(TestTensorflowSetup):
    def test_init(self):
        sess = self.sess
        env = make('Pendulum-v0')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        action_dim = env_spec.flat_action_dim
        state_dim = env_spec.flat_obs_dim
        # bs_shape = tf.placeholder(dtype=tf.int8, shape=[])
        bs_shape = 4
        action_ph = tf.placeholder(dtype=tf.float32, shape=[None, action_dim])
        state_ph = tf.placeholder(dtype=tf.float32, shape=[None, state_dim])

        mean_old = tf.layers.dense(inputs=state_ph,
                                   name='layer1',
                                   units=action_dim)

        mean2 = tf.layers.dense(inputs=state_ph,
                                name='layer2',
                                units=action_dim)

        # mean1 = tf.get_variable(name='mean1', shape=[bs_shape, action_dim], dtype=tf.float32)
        var1 = tf.get_variable(name='var1', shape=[action_dim], dtype=tf.float32,
                               initializer=tf.initializers.random_uniform(0.0, 1.0))

        # mean2 = tf.get_variable(name='mean2', shape=[bs_shape, action_dim], dtype=tf.float32)
        var2 = tf.get_variable(name='var2', shape=[action_dim], dtype=tf.float32,
                               initializer=tf.initializers.random_uniform(0.0, 1.0))

        # var1 = tf.get_variable('logvars', (10, action_dim), tf.float32,
        #                        tf.constant_initializer(0.0))
        # var1 = tf.expand_dims(tf.reduce_sum(var1, axis=0), axis=0)
        # var1 = tf.tile(var1, [bs_shape, 1])
        #
        # var2 = tf.get_variable('logvars2', (10, action_dim), tf.float32,
        #                        tf.constant_initializer(0.0))
        # var2 = tf.expand_dims(tf.reduce_sum(var2, axis=0), 0)
        # var2 = tf.tile(var2, [bs_shape, 1])

        dist_old = tfp.distributions.MultivariateNormalDiag(mean_old, tf.sqrt(var1), validate_args=True)
        dis2 = tfp.distributions.MultivariateNormalDiag(mean2, tf.sqrt(var2), validate_args=True)

        dist_norm1 = tfp.distributions.Normal(mean_old, var1)
        dist_norm2 = tfp.distributions.Normal(mean2, var2)

        print(dist_old, dis2)
        # dis1 = tfp.distributions.Independent(dis1, reinterpreted_batch_ndims=1)
        # dis2 = tfp.distributions.Independent(dis2, reinterpreted_batch_ndims=1)

        # op = tf.train.AdamOptimizer(learning_rate=0.1).minimize(tfp.distributions.kl_divergence(dis1, dis2),
        #                                                         var_list=[mean1, var1])

        ac = [env_spec.action_space.sample() for _ in range(bs_shape)]
        ac = make_batch(np.array(ac), original_shape=env_spec.action_shape)

        state = [env_spec.obs_space.sample() for _ in range(bs_shape)]
        state = make_batch(np.array(state), original_shape=env_spec.obs_shape)

        feed_dict = {
            state_ph: state,
            action_ph: ac
        }
        sess.run(tf.global_variables_initializer())

        kl, entropy, logp, log_p_old = kl_entropy_logprob_from_pat_cody(old_mean=mean_old,
                                                                        old_var=var1,
                                                                        mean=mean2,
                                                                        var=var2,
                                                                        feed_dict=feed_dict,
                                                                        sess=sess,
                                                                        action_ph=action_ph,
                                                                        action_dim=action_dim)

        kl_tfp = sess.run(tf.reduce_mean(tfp.distributions.kl_divergence(dist_old, dis2)), feed_dict=feed_dict)
        entropy_tfp = sess.run(tf.reduce_mean(dis2.entropy()), feed_dict=feed_dict)

        log_prob_tfp = sess.run(dis2.log_prob(value=ac), feed_dict=feed_dict)
        log_p_old_tfp = sess.run(dist_old.log_prob(value=ac), feed_dict=feed_dict)

        test_log_prob_tfp = dis2.log_prob(ac) + tf.cast(0.5 * np.log(2. * np.pi * action_dim), dtype=tf.float32)
        test_log_prob_tfp_old = dist_old.log_prob(ac) + tf.cast(0.5 * np.log(2. * np.pi * action_dim), dtype=tf.float32)

        print("ac shape {}".format(ac.shape))
        print("a sample from dis1 shape {}".format(sess.run(dist_old.sample(), feed_dict=feed_dict).shape))
        print("shape of dis under feeddict {}".format(
            sess.run([dist_old.batch_shape_tensor(), dist_old.event_shape_tensor()],
                     feed_dict=feed_dict)))

        # print(sess.run(dis2.log_prob(value=ac)).shape)
        # print(sess.run(dis1.log_prob(value=ac)).shape)
        for i in range(bs_shape):
            feed_dict_i = {
                state_ph: make_batch(state[i], env_spec.obs_shape),
                action_ph: make_batch(ac[i], env_spec.action_shape)
            }
            print("i dis2 log prob: {}".format(sess.run(dis2.log_prob(value=ac[i]), feed_dict=feed_dict_i)))
            print("i dis1 log prob: {}".format(sess.run(dist_old.log_prob(value=ac[i]), feed_dict=feed_dict_i)))

        print(kl, kl_tfp)
        print(entropy, entropy_tfp)

        print(logp, log_prob_tfp)
        print(log_p_old, log_p_old_tfp)
        print('new log p {}'.format(sess.run(test_log_prob_tfp, feed_dict=feed_dict)))
        print('new log p old {}'.format(sess.run(test_log_prob_tfp_old, feed_dict=feed_dict)))

        print('new log p norm {}'.format(sess.run(tf.reduce_sum(dist_norm1.log_prob(ac), axis=1), feed_dict=feed_dict)))
        print('new log p old norm {}'.format(
            sess.run(tf.reduce_sum(dist_norm2.log_prob(ac), axis=1), feed_dict=feed_dict)))

        self.assertTrue(np.isclose(logp, log_prob_tfp).all())
        self.assertTrue(np.isclose(log_p_old, log_p_old_tfp).all())
        self.assertTrue(np.isclose(kl, kl_tfp).all())
        self.assertTrue(np.isclose(entropy, entropy_tfp).all())

        kl, entropy, logp, log_p_old = kl_entropy_logprob_from_mvn(old_mean=mean_old,
                                                                   old_var=var1,
                                                                   mean=mean2,
                                                                   var=var2,
                                                                   feed_dict=feed_dict,
                                                                   sess=sess,
                                                                   action_ph=action_ph,
                                                                   action_dim=action_dim)
        print(kl, entropy, logp, log_p_old)
        self.assertTrue(np.isclose(logp, log_prob_tfp).all())
        self.assertTrue(np.isclose(log_p_old, log_p_old_tfp).all())
        self.assertTrue(np.isclose(entropy, entropy_tfp).all())
        self.assertTrue(np.isclose(kl, kl_tfp).all())



PENDULUM_BENCHMARK_CONFIG_DICT = {
    'env_id': "Pendulum-v0",
    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=10000,
                                         TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                         TOTAL_AGENT_UPDATE_COUNT=None),
    'DynaFlow': {
        "TEST_ALGO_EVERY_REAL_SAMPLE_COUNT": 200,
        "TEST_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 200,
        "TRAIN_DYNAMICS_EVERY_REAL_SAMPLE_COUNT": 10,
        "START_TRAIN_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TRAIN_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_ALGO_AFTER_SAMPLE_COUNT": 1,
        "START_TEST_DYNAMICS_AFTER_SAMPLE_COUNT": 1,
        "WARM_UP_DYNAMICS_SAMPLES": 2000,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_REAL_ENV": 10,
        "TRAIN_ALGO_EVERY_REAL_SAMPLE_COUNT_FROM_DYNAMICS_ENV": 50,
    },
    'DynamicsModel': dict(learning_rate=0.01,
                          mlp_config=[
                              {
                                  "ACT": "RELU",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "1",
                                  "N_UNITS": 32,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              },
                              {
                                  "ACT": "RELU",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "2",
                                  "N_UNITS": 64,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              },
                              {
                                  "ACT": "IDENTITY",
                                  "B_INIT_VALUE": 0.0,
                                  "NAME": "OUPTUT",
                                  "N_UNITS": 3,
                                  "TYPE": "DENSE",
                                  "W_NORMAL_STDDEV": 0.03
                              }
                          ]),
    'MPC': dict(
        SAMPLED_HORIZON=5,
        SAMPLED_PATH_NUM=20,
        dynamics_model_train_iter=10
    ),
}



"""
This script is from garage
"""

import gym.spaces
import numpy as np
from baconian.common import special
__all__ = [
    'flat_dim', 'flatten', 'flatten_n', 'unflatten', 'unflatten_n',
    'weighted_sample'
]


def flat_dim(space):
    if isinstance(space, gym.spaces.Box):
        return np.prod(space.low.shape)
    elif isinstance(space, gym.spaces.Discrete):
        return space.n
    elif isinstance(space, gym.spaces.Tuple):
        return np.sum([flat_dim(x) for x in space.spaces])
    else:
        raise NotImplementedError


def flatten(space, obs):
    if isinstance(space, gym.spaces.Box):
        return np.asarray(obs).flatten()
    elif isinstance(space, gym.spaces.Discrete):
        if space.n == 2:
            obs = int(obs)
        return special.to_onehot(obs, space.n)
    elif isinstance(space, gym.spaces.Tuple):
        return np.concatenate(
            [flatten(c, xi) for c, xi in zip(space.spaces, obs)])
    else:
        raise NotImplementedError


def flatten_n(space, obs):
    if isinstance(space, gym.spaces.Box):
        obs = np.asarray(obs)
        return obs.reshape((obs.shape[0], -1))
    elif isinstance(space, gym.spaces.Discrete):
        return special.to_onehot_n(obs, space.n)
    elif isinstance(space, gym.spaces.Tuple):
        obs_regrouped = [[obs[i] for o in obs] for i in range(len(obs[0]))]
        flat_regrouped = [
            flatten_n(c, oi) for c, oi in zip(space.spaces, obs_regrouped)
        ]
        return np.concatenate(flat_regrouped, axis=-1)
    else:
        raise NotImplementedError


def unflatten(space, obs):
    if isinstance(space, gym.spaces.Box):
        return np.asarray(obs).reshape(space.shape)
    elif isinstance(space, gym.spaces.Discrete):
        return special.from_onehot(obs)
    elif isinstance(space, gym.spaces.Tuple):
        dims = [flat_dim(c) for c in space.spaces]
        flat_xs = np.split(obs, np.cumsum(dims)[:-1])
        return tuple(unflatten(c, xi) for c, xi in zip(space.spaces, flat_xs))
    else:
        raise NotImplementedError


def unflatten_n(space, obs):
    if isinstance(space, gym.spaces.Box):
        obs = np.asarray(obs)
        return obs.reshape((obs.shape[0],) + space.shape)
    elif isinstance(space, gym.spaces.Discrete):
        return special.from_onehot_n(obs)
    elif isinstance(space, gym.spaces.Tuple):
        dims = [flat_dim(c) for c in space.spaces]
        flat_xs = np.split(obs, np.cumsum(dims)[:-1], axis=-1)
        unflat_xs = [
            unflatten_n(c, xi) for c, xi in zip(space.spaces, flat_xs)
        ]
        unflat_xs_grouped = list(zip(*unflat_xs))
        return unflat_xs_grouped
    else:
        raise NotImplementedError


def weighted_sample(space, weights):
    if isinstance(space, gym.spaces.Discrete):
        return special.weighted_sample(weights, range(space.n))
    else:
        raise NotImplementedError



from collections import Iterable
from collections import namedtuple

import numpy as np
import tensorflow as tf


def compile_function(inputs, outputs, log_name=None):
    def run(*input_vals):
        sess = tf.get_default_session()
        return sess.run(outputs, feed_dict=dict(list(zip(inputs, input_vals))))

    return run


def flatten_batch(t, name="flatten_batch"):
    return tf.reshape(t, [-1] + list(t.shape[2:]), name=name)


def flatten_batch_dict(d, name=None):
    with tf.name_scope(name, "flatten_batch_dict", [d]):
        return {k: flatten_batch(v) for k, v in d.items()}


def filter_valids(t, valid, name="filter_valids"):
    # 'valid' is either 0 or 1 with dtype of tf.float32
    # Must round before cast to prevent floating-error
    return tf.dynamic_partition(
        t, tf.to_int32(tf.round(valid)), 2, name=name)[1]


def filter_valids_dict(d, valid, name=None):
    with tf.name_scope(name, "filter_valids_dict", [d, valid]):
        return {k: filter_valids(v, valid) for k, v in d.items()}


def graph_inputs(name, **kwargs):
    Singleton = namedtuple(name, kwargs.keys())
    return Singleton(**kwargs)


def flatten_inputs(deep):
    def flatten(deep):
        for d in deep:
            if isinstance(d, Iterable) and not isinstance(
                    d, (str, bytes, tf.Tensor, np.ndarray)):
                yield from flatten(d)
            else:
                yield d

    return list(flatten(deep))


def flatten_tensor_variables(ts):
    return tf.concat(
        axis=0,
        values=[tf.reshape(x, [-1]) for x in ts],
        name="flatten_tensor_variables")


def unflatten_tensor_variables(flatarr, shapes, symb_arrs):
    arrs = []
    n = 0
    for (shape, symb_arr) in zip(shapes, symb_arrs):
        size = np.prod(list(shape))
        arr = tf.reshape(flatarr[n:n + size], shape)
        arrs.append(arr)
        n += size
    return arrs


def new_tensor(name, ndim, dtype):
    return tf.placeholder(dtype=dtype, shape=[None] * ndim, name=name)


def new_tensor_like(name, arr_like):
    return new_tensor(name,
                      arr_like.get_shape().ndims, arr_like.dtype.base_dtype)


def concat_tensor_list(tensor_list):
    return np.concatenate(tensor_list, axis=0)


def concat_tensor_dict_list(tensor_dict_list):
    keys = list(tensor_dict_list[0].keys())
    ret = dict()
    for k in keys:
        example = tensor_dict_list[0][k]
        if isinstance(example, dict):
            v = concat_tensor_dict_list([x[k] for x in tensor_dict_list])
        else:
            v = concat_tensor_list([x[k] for x in tensor_dict_list])
        ret[k] = v
    return ret


def stack_tensor_list(tensor_list):
    return np.array(tensor_list)
    # tensor_shape = np.array(tensor_list[0]).shape
    # if tensor_shape is tuple():
    #     return np.array(tensor_list)
    # return np.vstack(tensor_list)


def stack_tensor_dict_list(tensor_dict_list):
    """
    Stack a list of dictionaries of {tensors or dictionary of tensors}.
    :param tensor_dict_list: a list of dictionaries of {tensors or dictionary
     of tensors}.
    :return: a dictionary of {stacked tensors or dictionary of stacked tensors}
    """
    keys = list(tensor_dict_list[0].keys())
    ret = dict()
    for k in keys:
        example = tensor_dict_list[0][k]
        if isinstance(example, dict):
            v = stack_tensor_dict_list([x[k] for x in tensor_dict_list])
        else:
            v = stack_tensor_list([x[k] for x in tensor_dict_list])
        ret[k] = v
    return ret


def split_tensor_dict_list(tensor_dict):
    keys = list(tensor_dict.keys())
    ret = None
    for k in keys:
        vals = tensor_dict[k]
        if isinstance(vals, dict):
            vals = split_tensor_dict_list(vals)
        if ret is None:
            ret = [{k: v} for v in vals]
        else:
            for v, cur_dict in zip(vals, ret):
                cur_dict[k] = v
    return ret


def to_onehot_sym(inds, dim):
    return tf.one_hot(inds, depth=dim, on_value=1, off_value=0)


def pad_tensor(x, max_len):
    return np.concatenate([
        x,
        np.tile(
            np.zeros_like(x[0]), (max_len - len(x),) + (1,) * np.ndim(x[0]))
    ])


def pad_tensor_n(xs, max_len):
    ret = np.zeros((len(xs), max_len) + xs[0].shape[1:], dtype=xs[0].dtype)
    for idx, x in enumerate(xs):
        ret[idx][:len(x)] = x
    return ret


def pad_tensor_dict(tensor_dict, max_len):
    keys = list(tensor_dict.keys())
    ret = dict()
    for k in keys:
        if isinstance(tensor_dict[k], dict):
            ret[k] = pad_tensor_dict(tensor_dict[k], max_len)
        else:
            ret[k] = pad_tensor(tensor_dict[k], max_len)
    return ret


def compute_advantages(discount,
                       gae_lambda,
                       max_len,
                       baselines,
                       rewards,
                       name=None):
    with tf.name_scope(name, "compute_advantages",
                       [discount, gae_lambda, max_len, baselines, rewards]):
        # Calculate advantages
        #
        # Advantages are a discounted cumulative sum.
        #
        # The discount cumulative sum can be represented as an IIR
        # filter ob the reversed input vectors, i.e.
        #    y[t] - discount*y[t+1] = x[t]
        #        or
        #    rev(y)[t] - discount*rev(y)[t-1] = rev(x)[t]
        #
        # Given the time-domain IIR filter step response, we can
        # calculate the filter response to our signal by convolving the
        # signal with the filter response function. The time-domain IIR
        # step response is calculated below as discount_filter:
        #     discount_filter =
        #         [1, discount, discount^2, ..., discount^N-1]
        #         where the epsiode length is N.
        #
        # We convolve discount_filter with the reversed time-domain
        # signal deltas to calculate the reversed advantages:
        #     rev(advantages) = discount_filter (X) rev(deltas)
        #
        # TensorFlow's tf.nn.conv1d op is not a true convolution, but
        # actually a cross-correlation, so its input and output are
        # already implicitly reversed for us.
        #    advantages = discount_filter (tf.nn.conv1d) deltas

        # Prepare convolutional IIR filter to calculate advantages
        gamma_lambda = tf.constant(
            float(discount) * float(gae_lambda),
            dtype=tf.float32,
            shape=[max_len, 1, 1])
        advantage_filter = tf.cumprod(gamma_lambda, exclusive=True)

        # Calculate deltas
        pad = tf.zeros_like(baselines[:, :1])
        baseline_shift = tf.concat([baselines[:, 1:], pad], 1)
        deltas = rewards + discount * baseline_shift - baselines
        # Convolve deltas with the discount filter to get advantages
        deltas_pad = tf.expand_dims(
            tf.concat([deltas, tf.zeros_like(deltas[:, :-1])], axis=1), axis=2)
        adv = tf.nn.conv1d(
            deltas_pad, advantage_filter, stride=1, padding='VALID')
        advantages = tf.reshape(adv, [-1])
    return advantages


def discounted_returns(discount, max_len, rewards, name=None):
    with tf.name_scope(name, "discounted_returns",
                       [discount, max_len, rewards]):
        gamma = tf.constant(
            float(discount), dtype=tf.float32, shape=[max_len, 1, 1])
        return_filter = tf.cumprod(gamma, exclusive=True)
        rewards_pad = tf.expand_dims(
            tf.concat([rewards, tf.zeros_like(rewards[:, :-1])], axis=1),
            axis=2)
        returns = tf.nn.conv1d(
            rewards_pad, return_filter, stride=1, padding='VALID')
    return returns



from baconian.core.core import Basic, Env, EnvSpec
from baconian.envs.env_wrapper import Wrapper, ObservationWrapper, StepObservationWrapper
from baconian.common.sampler.sampler import Sampler
from baconian.common.error import *
from baconian.algo.algo import Algo
from typeguard import typechecked
from baconian.algo.misc import ExplorationStrategy
from baconian.common.sampler.sample_data import SampleData
from baconian.common.logging import Recorder, record_return_decorator
from baconian.core.status import StatusWithSubInfo
from baconian.core.status import register_counter_info_to_status_decorator
from baconian.core.util import init_func_arg_record_decorator
from baconian.common.logging import ConsoleLogger
from baconian.common.sampler.sample_data import TransitionData, TrajectoryData
from baconian.common.schedules import EventScheduler
from baconian.common.noise import AgentActionNoiseWrapper
from baconian.core.parameters import Parameters


class Agent(Basic):
    STATUS_LIST = ('CREATED', 'INITED', 'TRAIN', 'TEST')
    INIT_STATUS = 'CREATED'
    required_key_dict = {}

    @init_func_arg_record_decorator()
    @typechecked
    def __init__(self, name,
                 # config_or_config_dict: (DictConfig, dict),
                 env: (Env, Wrapper),
                 algo: Algo,
                 env_spec: EnvSpec,
                 sampler: Sampler = None,
                 noise_adder: AgentActionNoiseWrapper = None,
                 reset_noise_every_terminal_state=False,
                 reset_state_every_sample=False,
                 exploration_strategy: ExplorationStrategy = None,
                 algo_saving_scheduler: EventScheduler = None):
        """

        :param name: the name of the agent instance
        :type name: str
        :param env: environment that interacts with agent
        :type env: Env
        :param algo: algorithm of the agent
        :type algo: Algo
        :param env_spec: environment specifications: action apace and environment space
        :type env_spec: EnvSpec
        :param sampler: sampler
        :type sampler: Sampler
        :param reset_noise_every_terminal_state: reset the noise every sampled trajectory
        :type reset_noise_every_terminal_state: bool
        :param reset_state_every_sample: reset the state everytime perofrm the sample/rollout
        :type reset_state_every_sample: bool
        :param noise_adder: add action noise for exploration in action space
        :type noise_adder: AgentActionNoiseWrapper
        :param exploration_strategy: exploration strategy in action space
        :type exploration_strategy: ExplorationStrategy
        :param algo_saving_scheduler: control the schedule the varying parameters in training process
        :type algo_saving_scheduler: EventSchedule
        """
        super(Agent, self).__init__(name=name, status=StatusWithSubInfo(self))
        self.parameters = Parameters(parameters=dict(reset_noise_every_terminal_state=reset_noise_every_terminal_state,
                                                     reset_state_every_sample=reset_state_every_sample))
        self.env = env
        self.algo = algo
        self._env_step_count = 0
        if sampler is None:
            sampler = Sampler()
        self.sampler = sampler
        self.recorder = Recorder(default_obj=self)
        self.env_spec = env_spec
        if exploration_strategy:
            assert isinstance(exploration_strategy, ExplorationStrategy)
            self.explorations_strategy = exploration_strategy
        else:
            self.explorations_strategy = None
        self.noise_adder = noise_adder
        self.algo_saving_scheduler = algo_saving_scheduler

    # @record_return_decorator(which_recorder='self')
    @register_counter_info_to_status_decorator(increment=1, info_key='update_counter', under_status='TRAIN')
    def train(self, *args, **kwargs):
        """
        train the agent

        :return: True for successfully train the agent, false if memory buffer did not have enough data.
        :rtype: bool
        """
        self.set_status('TRAIN')
        self.algo.set_status('TRAIN')
        ConsoleLogger().print('info', 'train agent:')
        try:
            res = self.algo.train(*args, **kwargs)
        except MemoryBufferLessThanBatchSizeError as e:
            ConsoleLogger().print('warning', 'memory buffer did not have enough data to train, skip training')
            return False

        ConsoleLogger().print('info', res)

        if self.algo_saving_scheduler and self.algo_saving_scheduler.value() is True:
            self.algo.save(global_step=self._status.get_specific_info_key_status(info_key='update_counter',
                                                                                 under_status='TRAIN'))

    # @record_return_decorator(which_recorder='self')
    def test(self, sample_count) -> SampleData:
        """

        test the agent

        :param sample_count: how many trajectories used to evaluate the agent's performance
        :type sample_count: int
        :return: SampleData object.
        """
        self.set_status('TEST')
        self.algo.set_status('TEST')
        ConsoleLogger().print('info', 'test: agent with {} trajectories'.format(sample_count))
        res = self.sample(env=self.env,
                          sample_count=sample_count,
                          sample_type='trajectory',
                          store_flag=False,
                          in_which_status='TEST')
        return res

    @register_counter_info_to_status_decorator(increment=1, info_key='predict_counter', under_status=('TRAIN', 'TEST'),
                                               ignore_wrong_status=True)
    def predict(self, **kwargs):
        """
        predict the action given the state

        :param kwargs: rest parameters, include key: obs
        :return: predicted action
        :rtype: numpy ndarray
        """
        res = None
        if self.explorations_strategy and not self.is_testing:
            res = self.explorations_strategy.predict(**kwargs, algo=self.algo)
        else:
            if self.noise_adder and not self.is_testing:
                res = self.env_spec.action_space.clip(self.noise_adder(self.algo.predict(**kwargs)))
            else:
                res = self.algo.predict(**kwargs)
        self.recorder.append_to_obj_log(obj=self, attr_name='action', status_info=self.get_status(), log_val=res)
        return res

    @register_counter_info_to_status_decorator(increment=1, info_key='sample_counter', under_status=('TRAIN', 'TEST'),
                                               ignore_wrong_status=True)
    def sample(self, env, sample_count: int, in_which_status: str = 'TRAIN', store_flag=False,
               sample_type: str = 'transition') -> (
            TransitionData, TrajectoryData):
        """
        sample a certain number of data from environment

        :param env: environment to sample
        :param sample_count: int, sample count
        :param in_which_status: string, environment status
        :param store_flag: to store environment samples or not, default False
        :param sample_type: the type of sample, 'transition' by default
        :return: sample data from environment
        :rtype: some subclass of SampleData: TrajectoryData or TransitionData
        """
        self.set_status(in_which_status)
        env.set_status(in_which_status)
        self.algo.set_status(in_which_status)
        ConsoleLogger().print('info',
                              "agent sampled {} {} under status {}".format(sample_count, sample_type,
                                                                           self.get_status()))
        batch_data = self.sampler.sample(agent=self,
                                         env=env,
                                         reset_at_start=self.parameters('reset_state_every_sample'),
                                         sample_type=sample_type,
                                         sample_count=sample_count)
        if store_flag is True:
            self.store_samples(samples=batch_data)
        # todo when we have transition/ trajectory data here, the mean or sum results are still valid?
        ConsoleLogger().print('info',
                              "sample: mean reward {}, sum reward {}".format(
                                  batch_data.get_mean_of(set_name='reward_set'),
                                  batch_data.get_sum_of(set_name='reward_set')))
        self.recorder.append_to_obj_log(obj=self, attr_name='average_reward',
                                        status_info=self.get_status(),
                                        log_val=batch_data.get_mean_of('reward_set'))
        self.recorder.append_to_obj_log(obj=self, attr_name='sum_reward',
                                        status_info=self.get_status(),
                                        log_val=batch_data.get_sum_of('reward_set'))
        return batch_data

    def reset_on_terminal_state(self):
        if self.parameters('reset_noise_every_terminal_state') is True and self.noise_adder is not None:
            self.noise_adder.reset()

    def init(self):
        """
        Initialize the algorithm, and set status to 'INITED'.
        """
        self.algo.init()
        self.set_status('INITED')
        self.algo.warm_up(trajectory_data=self.sampler.sample(env=self.env,
                                                              agent=self,
                                                              sample_type='trajectory',
                                                              reset_at_start=True,
                                                              sample_count=self.algo.warm_up_trajectories_number))

    @typechecked
    def store_samples(self, samples: SampleData):
        """
        store the samples into memory/replay buffer if the algorithm that agent hold need to do so, like DQN, DDPG

        :param samples: sample data of the experiment
        :type samples: SampleData
        """
        self.algo.append_to_memory(samples=samples)

    @property
    def is_training(self):
        """
        Check whether the agent is training. Return a boolean value.

        :return: true if the agent is training
        :rtype: bool
        """
        return self.get_status()['status'] == 'TRAIN'

    @property
    def is_testing(self):
        """
        Check whether the agent is testing. Return a boolean value.

        :return: true if the agent is testing
        :rtype: bool
        """
        return self.get_status()['status'] == 'TEST'



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.algo.dynamics.reward_func.reward_func import QuadraticCostFunc
from baconian.envs.gym_env import make
import numpy as np
from baconian.core.core import EnvSpec
from baconian.algo.dynamics.linear_dynamics_model import LinearDynamicsModel
from baconian.algo.policy.lqr_policy import LQRPolicy


class TestLQRPolicy(TestWithAll):
    default_id = -1

    def test_correctness(self):
        env_id = 'Pendulum-v0'
        env = make(env_id)
        n = env.observation_space.flat_dim
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        F = np.ones([env.observation_space.flat_dim,
                     env.observation_space.flat_dim + env.action_space.flat_dim]) * 0.00001
        # F[n:, n:] = 0.0001
        dyna = LinearDynamicsModel(env_spec=env_spec,
                                   state_transition_matrix=F,
                                   bias=np.zeros([env.observation_space.flat_dim]))
        C = np.ones([env.observation_space.flat_dim + env.action_space.flat_dim,
                     env.observation_space.flat_dim + env.action_space.flat_dim]) * 0.00001
        c = np.ones([env.observation_space.flat_dim + env.action_space.flat_dim])
        c[n:] = -1000
        # C[:n, :] = 0.
        # C[:, :n] = 0.
        # c[:n] = 0.0
        cost_fn = QuadraticCostFunc(C=C, c=c)

        policy = LQRPolicy(env_spec=env_spec,
                           T=5,
                           dynamics=dyna,
                           cost_fn=cost_fn)
        st = env.reset() * 0.0
        for i in range(10):
            ac = policy.forward(st)
            st = dyna.step(action=ac, state=st, allow_clip=True)
            print(cost_fn(state=st, action=ac, new_state=None))
            print(st, ac)



from baconian.common.sampler.sample_data import TransitionData, TrajectoryData
from baconian.algo.value_func import ValueFunction
import scipy.signal
from baconian.common.special import *


def discount(x, gamma):
    """code clip from pat-cody"""
    """ Calculate discounted forward sum of a sequence at each point """
    return scipy.signal.lfilter([1.0], [1.0, -gamma], x[::-1])[::-1]


class SampleProcessor(object):

    @staticmethod
    def add_gae(data: TrajectoryData, gamma, lam, value_func: ValueFunction = None, name='advantage_set'):
        for traj in data.trajectories:
            # scale if gamma less than 1
            rewards = traj('reward_set') * (1 - gamma) if gamma < 0.999 else traj('reward_set')
            try:
                traj('v_value_set')
            except ValueError:
                if value_func is None:
                    raise ValueError('v_value_set did not existed, pass in value_func parameter to compute v_value_set')
                SampleProcessor.add_estimated_v_value(
                    data=traj,
                    value_func=value_func
                )
            finally:
                values = traj('v_value_set')
            # todo better way to handle shape error (no squeeze)
            tds = np.squeeze(rewards) - np.squeeze(values) + np.append(values[1:] * gamma, 0)
            advantages = discount(tds, gamma * lam)
            traj.append_new_set(name=name, data_set=make_batch(advantages, original_shape=[]), shape=[])

    @staticmethod
    def add_discount_sum_reward(data: TrajectoryData, gamma, name='discount_set'):
        for traj in data.trajectories:
            # scale if gamma less than 1
            dis_set = traj('reward_set') * (1 - gamma) if gamma < 0.999 else traj('reward_set')
            # TODO add a unit test
            dis_reward_set = discount(np.reshape(dis_set, [-1, ]), gamma)
            traj.append_new_set(name=name, data_set=make_batch(dis_reward_set, original_shape=[]), shape=[])

    @staticmethod
    def add_estimated_v_value(data: (TrajectoryData, TransitionData), value_func: ValueFunction, name='v_value_set'):
        if isinstance(data, TrajectoryData):
            for path in data.trajectories:
                SampleProcessor._add_estimated_v_value(path, value_func, name)
        else:
            SampleProcessor._add_estimated_v_value(data, value_func, name)

    @staticmethod
    def _add_estimated_v_value(data: TransitionData, value_func: ValueFunction, name):
        v_set = value_func.forward(data.state_set)
        data.append_new_set(name=name, data_set=make_batch(np.array(v_set), original_shape=[]), shape=[])

    @staticmethod
    def normalization(data: (TransitionData, TrajectoryData), key, mean: np.ndarray = None, std_dev: np.ndarray = None):
        if isinstance(data, TransitionData):
            if mean is not None:
                assert mean.shape == data(key).shape[1:]
                assert std_dev.shape == data(key).shape[1:]
            else:
                mean = data(key).mean(axis=0)
                std_dev = data(key).std(axis=0)
            data.append_new_set(name=key, data_set=(data(key) - mean) / (std_dev + 1e-6), shape=data(key).shape[1:])
            return data
        else:
            # TODO add shape check
            mean = np.mean([d(key) for d in data.trajectories])
            std_dev = np.std([d(key) for d in data.trajectories])
            for d in data.trajectories:
                d.append_new_set(name=key, data_set=np.array((d(key) - mean) / (std_dev + 1e-6)),
                                 shape=d(key).shape[1:])
            return data



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.common.sampler.sample_data import TransitionData
from baconian.algo.policy import UniformRandomPolicy
from baconian.algo.dynamics.gaussian_mixture_dynamics_prior import GaussianMixtureDynamicsPrior


class TestDynamicsPrior(TestWithAll):
	def test_update(self):
		env = make('Pendulum-v0')
		name = 'demo_exp'
		env_spec = EnvSpec(obs_space=env.observation_space, action_space=env.action_space)
		data = TransitionData(env_spec=env_spec)
		policy = UniformRandomPolicy(env_spec=env_spec)

		# Do some initial sampling here to train gmm model
		st = env.reset()
		for i in range(100):
			ac = policy.forward(st)
			new_st, re, _, _ = env.step(ac)
			data.append(state=st, new_state=new_st, action=ac, reward=re, done=False)
			st = new_st

		gmm = GaussianMixtureDynamicsPrior(env_spec=env_spec, batch_data=data)
		gmm.init()
		gmm.update(batch_data=data)

	def test_prior_eval(self):
		env = make('Pendulum-v0')
		name = 'demo_exp'
		env_spec = EnvSpec(obs_space=env.observation_space, action_space=env.action_space)
		data = TransitionData(env_spec=env_spec)
		policy = UniformRandomPolicy(env_spec=env_spec)

		# Do some initial sampling here to train gmm model
		st = env.reset()
		for i in range(100):
			ac = policy.forward(st)
			new_st, re, _, _ = env.step(ac)
			data.append(state=st, new_state=new_st, action=ac, reward=re, done=False)
			st = new_st

		gmm = GaussianMixtureDynamicsPrior(env_spec=env_spec, batch_data=data)
		gmm.init()
		gmm.update(batch_data=data)
		mu0, Phi, m, n0 = gmm.eval(batch_data=data)

		state_shape = data.state_set.shape[1]
		action_shape = data.action_set.shape[1]
		self.assertEqual(state_shape + action_shape + state_shape, mu0.shape[0])
		self.assertEqual(state_shape + action_shape + state_shape, Phi.shape[0])
		self.assertEqual(state_shape + action_shape + state_shape, Phi.shape[1])



PENDULUM_BENCHMARK_CONFIG_DICT = {
    'env_id': "Pendulum-v0",
    'MLP_V': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": 30,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": 12,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": 5,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ]
    },
    'POLICY': {
        'mlp_config': [
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "N_UNITS": 30,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "2",
                "N_UNITS": 17,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "TANH",
                "B_INIT_VALUE": 0.0,
                "NAME": "3",
                "N_UNITS": 10,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "IDENTITY",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "N_UNITS": 1,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ]
    },
    'PPO': {
        'config_or_config_dict': {
            "gamma": 0.99,
            "lam": 0.98,
            "policy_train_iter": 20,
            "value_func_train_iter": 10,
            "clipping_range": [0.2, 0.2],
            "beta": 1.0,
            "eta": 50,
            "log_var_init": -1.0,
            "kl_target": 0.03,
            "policy_lr": 0.0003,
            "value_func_lr": 0.005,
            "value_func_train_batch_size": 32,
            "lr_multiplier": 1.0
        }
    },
    'TrainTestFlow': {
        "TEST_SAMPLES_COUNT": 1,
        "TRAIN_SAMPLES_COUNT": 32,
        'config_or_config_dict': {
            "TEST_EVERY_SAMPLE_COUNT": 1000,
            "TRAIN_EVERY_SAMPLE_COUNT": 10,
            "START_TRAIN_AFTER_SAMPLE_COUNT": 0,
            "START_TEST_AFTER_SAMPLE_COUNT": 0,
        }
    },
    'DEFAULT_EXPERIMENT_END_POINT': dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=1000 * 10,
                                         TOTAL_AGENT_TEST_SAMPLE_COUNT=None,
                                         TOTAL_AGENT_UPDATE_COUNT=None),
}



import numpy as np

from baconian.common.spaces.base import Space


class Tuple(Space):
    def __init__(self, *components):
        if isinstance(components[0], (list, tuple)):
            assert len(components) == 1
            components = components[0]
        self._components = tuple(components)
        dtypes = [
            c.new_tensor_variable("tmp", extra_dims=0).dtype
            for c in components
        ]
        if dtypes and hasattr(dtypes[0], "as_numpy_dtype"):
            dtypes = [d.as_numpy_dtype for d in dtypes]
        self._common_dtype = np.core.numerictypes.find_common_type([], dtypes)

    def sample(self):
        return tuple(x.sample() for x in self._components)

    @property
    def components(self):
        return self._components

    def contains(self, x):
        return bool(isinstance(x, tuple) and all(
            c.contains(xi) for c, xi in zip(self._components, x)))

    @property
    def flat_dim(self):
        return np.sum([c.flat_dim for c in self._components])

    def flatten(self, x):
        return np.concatenate(
            [c.flatten(xi) for c, xi in zip(self._components, x)])

    def flatten_n(self, xs):
        xs_regrouped = [[x[i] for x in xs] for i in range(len(xs[0]))]
        flat_regrouped = [
            c.flatten_n(xi) for c, xi in zip(self.components, xs_regrouped)
        ]
        return np.concatenate(flat_regrouped, axis=-1)

    def unflatten(self, x):
        dims = [c.flat_dim for c in self._components]
        flat_xs = np.split(x, np.cumsum(dims)[:-1])
        return tuple(
            c.unflatten(xi) for c, xi in zip(self._components, flat_xs))

    def unflatten_n(self, xs):
        dims = [c.flat_dim for c in self._components]
        flat_xs = np.split(xs, np.cumsum(dims)[:-1], axis=-1)
        unflat_xs = [
            c.unflatten_n(xi) for c, xi in zip(self.components, flat_xs)
        ]
        unflat_xs_grouped = list(zip(*unflat_xs))
        return unflat_xs_grouped

    def __eq__(self, other):
        if not isinstance(other, Tuple):
            return False
        return tuple(self.components) == tuple(other.components)

    def __hash__(self):
        return hash(tuple(self.components))

    def new_tensor_variable(self, name, extra_dims):
        raise NotImplementedError

    def bound(self):
        return (c.bound for c in self._components)



import unittest
from baconian.envs.gym_env import make
from baconian.core.core import EnvSpec
from baconian.algo.value_func import MLPVValueFunc
from baconian.test.tests.set_up.setup import TestTensorflowSetup
import tensorflow as tf


class TestMLPVValueFunc(TestTensorflowSetup):
    def test_init(self):
        env = make('Pendulum-v0')

        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        state_input = tf.placeholder(shape=[None, env_spec.flat_obs_dim],
                                     dtype=tf.float32,
                                     name='state_ph')

        mlp_v = MLPVValueFunc(env_spec=env_spec,
                              name_scope='mlp_q',
                              name='mlp_q',
                              state_input=state_input,
                              output_low=None,
                              output_high=None,
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
        mlp_v.init()
        mlp_v.forward(obs=env.observation_space.sample())

    def test_copy(self):
        env = make('Pendulum-v0')

        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        state_input = tf.placeholder(shape=[None, env_spec.flat_obs_dim],
                                     dtype=tf.float32,
                                     name='state_ph')

        mlp_v = MLPVValueFunc(env_spec=env_spec,
                              name_scope='mlp_v',
                              name='mlp_v',
                              state_input=state_input,
                              output_low=None,
                              output_high=None,
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
        mlp_v.init()

        new_mlp = mlp_v.make_copy(name='new_mlp',
                                  name_scope='mlp_v',
                                  reuse=True)

        new_mlp.init()

        self.assertGreater(len(mlp_v.parameters('tf_var_list')), 0)
        self.assertGreater(len(new_mlp.parameters('tf_var_list')), 0)

        for var1, var2 in zip(mlp_v.parameters('tf_var_list'), new_mlp.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertEqual(id(var1), id(var2))

        not_reuse_mlp = mlp_v.make_copy(name='no-reuse-mlp',
                                        name_scope='mlp_no_reuse',
                                        reuse=False)
        not_reuse_mlp.init()
        self.assertGreater(len(not_reuse_mlp.parameters('tf_var_list')), 0)

        for var1, var2 in zip(mlp_v.parameters('tf_var_list'), not_reuse_mlp.parameters('tf_var_list')):
            self.assertEqual(var1.shape, var2.shape)
            self.assertNotEqual(id(var1), id(var2))


if __name__ == '__main__':
    unittest.main()



from baconian.test.tests.set_up.setup import TestWithAll
from baconian.config.global_config import GlobalConfig

from baconian.algo.dqn import DQN
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.common.sampler.sample_data import TransitionData
import numpy as np


class TestDQN(TestWithAll):
    def test_init(self):
        dqn, locals = self.create_dqn()
        env = locals['env']
        env_spec = locals['env_spec']
        dqn.init()
        st = env.reset()
        a = TransitionData(env_spec)
        for i in range(100):
            ac = dqn.predict(obs=st, sess=self.sess, batch_flag=False)
            st_new, re, done, _ = env.step(action=ac)
            a.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
            st = st_new
            dqn.append_to_memory(a)
        new_dqn, _ = self.create_dqn(name='new_dqn')
        new_dqn.copy_from(dqn)
        self.assert_var_list_id_no_equal(dqn.q_value_func.parameters('tf_var_list'),
                                         new_dqn.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_id_no_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                         new_dqn.target_q_value_func.parameters('tf_var_list'))

        self.assert_var_list_equal(dqn.q_value_func.parameters('tf_var_list'),
                                   new_dqn.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                   new_dqn.target_q_value_func.parameters('tf_var_list'))

        dqn.save(save_path=GlobalConfig().DEFAULT_LOG_PATH + '/dqn_test',
                 global_step=0,
                 name=dqn.name)

        for i in range(10):
            print(dqn.train(batch_data=a, train_iter=10, sess=None, update_target=True))
            print(dqn.train(batch_data=None, train_iter=10, sess=None, update_target=True))

        self.assert_var_list_at_least_not_equal(dqn.q_value_func.parameters('tf_var_list'),
                                                new_dqn.q_value_func.parameters('tf_var_list'))

        self.assert_var_list_at_least_not_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                                new_dqn.target_q_value_func.parameters('tf_var_list'))

        dqn.load(path_to_model=GlobalConfig().DEFAULT_LOG_PATH + '/dqn_test',
                 model_name=dqn.name,
                 global_step=0)

        self.assert_var_list_equal(dqn.q_value_func.parameters('tf_var_list'),
                                   new_dqn.q_value_func.parameters('tf_var_list'))
        self.assert_var_list_equal(dqn.target_q_value_func.parameters('tf_var_list'),
                                   new_dqn.target_q_value_func.parameters('tf_var_list'))
        for i in range(10):
            self.sess.run(dqn.update_target_q_value_func_op,
                          feed_dict=dqn.parameters.return_tf_parameter_feed_dict())
            var1 = self.sess.run(dqn.q_value_func.parameters('tf_var_list'))
            var2 = self.sess.run(dqn.target_q_value_func.parameters('tf_var_list'))
            import numpy as np
            total_diff = 0.0
            for v1, v2 in zip(var1, var2):
                total_diff += np.mean(np.abs(np.array(v1) - np.array(v2)))
            print('update target, difference mean', total_diff)

    def test_l1_l2_norm(self):
        env = make('Acrobot-v1')
        env_spec = EnvSpec(obs_space=env.observation_space,
                           action_space=env.action_space)
        name = 'dqn'

        mlp_q = MLPQValueFunction(env_spec=env_spec,
                                  name_scope=name + '_mlp',
                                  name=name + '_mlp',
                                  mlp_config=[
                                      {
                                          "ACT": "RELU",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "1",
                                          "N_UNITS": 16,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03,
                                          "L1_NORM": 1000.0,
                                          "L2_NORM": 1000.0
                                      },
                                      {
                                          "ACT": "LINEAR",
                                          "B_INIT_VALUE": 0.0,
                                          "NAME": "OUPTUT",
                                          "N_UNITS": 1,
                                          "L1_NORM": 1000.0,
                                          "L2_NORM": 1000.0,
                                          "TYPE": "DENSE",
                                          "W_NORMAL_STDDEV": 0.03
                                      }
                                  ])
        dqn = DQN(env_spec=env_spec,
                  config_or_config_dict=dict(REPLAY_BUFFER_SIZE=1000,
                                             GAMMA=0.99,
                                             BATCH_SIZE=10,
                                             LEARNING_RATE=0.01,
                                             TRAIN_ITERATION=1,
                                             DECAY=0.5),
                  name=name,
                  value_func=mlp_q)
        dqn2, _ = self.create_dqn(name='dqn_2')
        a = TransitionData(env_spec)
        st = env.reset()
        dqn.init()
        dqn2.init()
        for i in range(100):
            ac = dqn.predict(obs=st, sess=self.sess, batch_flag=False)
            st_new, re, done, _ = env.step(action=ac)
            a.append(state=st, new_state=st_new, action=ac, done=done, reward=re)
            st = st_new
            dqn.append_to_memory(a)
        for i in range(20):
            print('dqn1 loss: ', dqn.train(batch_data=a, train_iter=10, sess=None, update_target=True))
            print('dqn2 loss: ', dqn2.train(batch_data=a, train_iter=10, sess=None, update_target=True))
        var_list = self.sess.run(dqn.q_value_func.parameters('tf_var_list'))
        print(var_list)
        var_list2 = self.sess.run(dqn2.q_value_func.parameters('tf_var_list'))
        print(var_list2)
        for var, var2 in zip(var_list, var_list2):
            diff = np.abs(var2) - np.abs(var)
            self.assertTrue(np.greater(np.mean(diff), 0.0).all())



"""
MPC benchmark on Pendulum
"""

from baconian.benchmark.mpc_benchmark.pendulum_conf import *
from baconian.algo.dynamics.terminal_func.terminal_func import FixedEpisodeLengthTerminalFunc
from baconian.core.flow.dyna_flow import DynaFlow
from baconian.envs.envs_reward_func import REWARD_FUNC_DICT
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.core.agent import Agent
from baconian.core.experiment import Experiment
from baconian.algo.mpc import ModelPredictiveControl
from baconian.algo.policy import UniformRandomPolicy
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect


def pendulum_task_fn():
    exp_config = PENDULUM_BENCHMARK_CONFIG_DICT
    GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT',
                       exp_config['DEFAULT_EXPERIMENT_END_POINT'])

    env = make('Pendulum-v0')
    name = 'benchmark'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_dyna = ContinuousMLPGlobalDynamicsModel(
        env_spec=env_spec,
        name_scope=name + '_mlp_dyna',
        name=name + '_mlp_dyna',
        **exp_config['DynamicsModel']
    )
    algo = ModelPredictiveControl(
        dynamics_model=mlp_dyna,
        env_spec=env_spec,
        config_or_config_dict=exp_config['MPC'],
        name=name + '_mpc',
        policy=UniformRandomPolicy(env_spec=env_spec, name='uni_policy')
    )
    algo.set_terminal_reward_function_for_dynamics_env(reward_func=REWARD_FUNC_DICT['Pendulum-v0'](),
                                                       terminal_func=FixedEpisodeLengthTerminalFunc(
                                                           max_step_length=env.unwrapped._max_episode_steps,
                                                           step_count_fn=algo.dynamics_env.total_step_count_fn), )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=algo,
                  exploration_strategy=None,
                  noise_adder=None,
                  name=name + '_agent')

    flow = DynaFlow(
        train_sample_count_func=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
        config_or_config_dict=exp_config['DynaFlow'],
        func_dict={
            'train_dynamics': {'func': agent.train,
                               'args': list(),
                               'kwargs': dict()},
            'train_algo': None,
            'test_algo': {'func': agent.test,
                          'args': list(),
                          'kwargs': dict(sample_count=1)},
            'test_dynamics': {'func': agent.algo.test_dynamics,
                              'args': list(),
                              'kwargs': dict(sample_count=100, env=env)},
            'sample_from_real_env': {'func': agent.sample,
                                     'args': list(),
                                     'kwargs': dict(sample_count=10,
                                                    env=agent.env,
                                                    in_which_status='TRAIN',
                                                    store_flag=True)},
            'sample_from_dynamics_env': None,
            'train_algo_from_synthesized_data': None
        }
    )

    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()



"""
Code from https://github.com/nrontsis/PILCO
"""

import tensorflow as tf
import gpflow
import numpy as np
from typeguard import typechecked

float_type = gpflow.settings.dtypes.float_type


def randomize(model):
    mean = 1
    sigma = 0.01

    model.kern.lengthscales.assign(
        mean + sigma * np.random.normal(size=model.kern.lengthscales.shape))
    model.kern.variance.assign(
        mean + sigma * np.random.normal(size=model.kern.variance.shape))
    if model.likelihood.variance.trainable:
        model.likelihood.variance.assign(
            mean + sigma * np.random.normal())


class MGPR(gpflow.Parameterized):
    def __init__(self, action_dim, state_dim, x, y, name=None):
        super(MGPR, self).__init__(name)

        self.num_outputs = state_dim
        self.num_dims = action_dim + state_dim
        self.optimizers = []
        self.models = []
        self._create_models(X=x, Y=y)

    def _create_models(self, X: np.ndarray, Y: np.ndarray):
        for i in range(self.num_outputs):
            kern = gpflow.kernels.RBF(input_dim=X.shape[1], ARD=True)
            kern.lengthscales.prior = gpflow.priors.Gamma(1, 10)  # priors have to be included before
            kern.variance.prior = gpflow.priors.Gamma(1.5, 2)  # before the model gets compiled
            # TODO: Maybe fix noise for better conditioning
            model = gpflow.models.GPR(X, Y[:, i:i + 1], kern, name="{}_{}".format(self.name, i))
            model.clear()
            model.compile()
            self.models.append(model)

    def set_XY(self, X: np.ndarray, Y: np.ndarray):
        for i in range(len(self.models)):
            self.models[i].X = X
            self.models[i].Y = Y[:, i:i + 1]

    def optimize(self, restarts=1):
        if len(self.optimizers) == 0:  # This is the first call to optimize();
            for model in self.models:
                # Create an gpflow.train.ScipyOptimizer object for every model embedded in mgpr
                optimizer = gpflow.train.ScipyOptimizer(method='L-BFGS-B')
                optimizer.minimize(model)
                self.optimizers.append(optimizer)
                restarts -= 1

        for model, optimizer in zip(self.models, self.optimizers):
            session = optimizer._model.enquire_session(None)
            best_parameters = model.read_values(session=session)
            best_likelihood = model.compute_log_likelihood()
            for restart in range(restarts):
                randomize(model)
                optimizer._optimizer.minimize(session=session,
                                              feed_dict=optimizer._gen_feed_dict(optimizer._model, None),
                                              step_callback=None)
                likelihood = model.compute_log_likelihood()
                if likelihood > best_likelihood:
                    best_parameters = model.read_values(session=session)
                    best_likelihood = likelihood
            model.assign(best_parameters)

    def predict(self, x):
        means = []
        vars = []
        for i in range(self.num_outputs):
            mean, var = self.models[i].predict_f(x)
            means.append(mean)
            vars.append(var)
        return np.array(means), np.array(vars)

    def predict_on_noisy_inputs(self, m, s):
        iK, beta = self.calculate_factorizations()
        return self.predict_given_factorizations(m, s, iK, beta)

    def calculate_factorizations(self):
        K = self.K(self.X)
        batched_eye = tf.eye(tf.shape(self.X)[0], batch_shape=[self.num_outputs], dtype=float_type)
        L = tf.cholesky(K + self.noise[:, None, None] * batched_eye)
        iK = tf.cholesky_solve(L, batched_eye)
        Y_ = tf.transpose(self.Y)[:, :, None]
        # Why do we transpose Y? Maybe we need to change the definition of self.Y() or beta?
        beta = tf.cholesky_solve(L, Y_)[:, :, 0]
        return iK, beta

    def predict_given_factorizations(self, m, s, iK, beta):
        """
        Approximate GP regression at noisy inputs via moment matching
        IN: mean (m) (row vector) and (s) variance of the state
        OUT: mean (M) (row vector), variance (S) of the action
             and inv(s)*input-ouputcovariance
        """

        s = tf.tile(s[None, None, :, :], [self.num_outputs, self.num_outputs, 1, 1])
        inp = tf.tile(self.centralized_input(m)[None, :, :], [self.num_outputs, 1, 1])

        # Calculate M and V: mean and inv(s) times input-output covariance
        iL = tf.matrix_diag(1 / self.lengthscales)
        iN = inp @ iL
        B = iL @ s[0, ...] @ iL + tf.eye(self.num_dims, dtype=float_type)

        # Redefine iN as in^T and t --> t^T
        # B is symmetric so its the same
        t = tf.linalg.transpose(
            tf.matrix_solve(B, tf.linalg.transpose(iN), adjoint=True),
        )

        lb = tf.exp(-tf.reduce_sum(iN * t, -1) / 2) * beta
        tiL = t @ iL
        c = self.variance / tf.sqrt(tf.linalg.det(B))

        M = (tf.reduce_sum(lb, -1) * c)[:, None]
        V = tf.matmul(tiL, lb[:, :, None], adjoint_a=True)[..., 0] * c[:, None]

        # Calculate S: Predictive Covariance
        R = s @ tf.matrix_diag(
            1 / tf.square(self.lengthscales[None, :, :]) +
            1 / tf.square(self.lengthscales[:, None, :])
        ) + tf.eye(self.num_dims, dtype=float_type)

        # TODO: change this block according to the PR of tensorflow. Maybe move it into a function?
        X = inp[None, :, :, :] / tf.square(self.lengthscales[:, None, None, :])
        X2 = -inp[:, None, :, :] / tf.square(self.lengthscales[None, :, None, :])
        Q = tf.matrix_solve(R, s) / 2
        Xs = tf.reduce_sum(X @ Q * X, -1)
        X2s = tf.reduce_sum(X2 @ Q * X2, -1)
        maha = -2 * tf.matmul(X @ Q, X2, adjoint_b=True) + \
               Xs[:, :, :, None] + X2s[:, :, None, :]
        #
        k = tf.log(self.variance)[:, None] - \
            tf.reduce_sum(tf.square(iN), -1) / 2
        L = tf.exp(k[:, None, :, None] + k[None, :, None, :] + maha)
        S = (tf.tile(beta[:, None, None, :], [1, self.num_outputs, 1, 1])
             @ L @
             tf.tile(beta[None, :, :, None], [self.num_outputs, 1, 1, 1])
             )[:, :, 0, 0]

        diagL = tf.transpose(tf.linalg.diag_part(tf.transpose(L)))
        S = S - tf.diag(tf.reduce_sum(tf.multiply(iK, diagL), [1, 2]))
        S = S / tf.sqrt(tf.linalg.det(R))
        S = S + tf.diag(self.variance)
        S = S - M @ tf.transpose(M)

        return tf.transpose(M), S, tf.transpose(V)

    def centralized_input(self, m):
        return self.X - m

    def K(self, X1, X2=None):
        return tf.stack(
            [model.kern.K(X1, X2) for model in self.models]
        )

    @property
    def Y(self):
        return tf.concat(
            [model.Y.parameter_tensor for model in self.models],
            axis=1
        )

    @property
    def X(self):
        return self.models[0].X.parameter_tensor

    @property
    def lengthscales(self):
        return tf.stack(
            [model.kern.lengthscales.constrained_tensor for model in self.models]
        )

    @property
    def variance(self):
        return tf.stack(
            [model.kern.variance.constrained_tensor for model in self.models]
        )

    @property
    def noise(self):
        return tf.stack(
            [model.likelihood.variance.constrained_tensor for model in self.models]
        )



from baconian.core.core import Basic
import abc
import numpy as np


class RewardFunc(Basic):
    allow_duplicate_name = True

    def __init__(self, name='reward_func'):
        super().__init__(name=name)

    @abc.abstractmethod
    def __call__(self, state, action, new_state, **kwargs) -> float:
        raise NotImplementedError

    def init(self):
        pass


class RandomRewardFunc(RewardFunc):
    """
    Debug and test use only
    """

    def __init__(self, name='random_reward_func'):
        super().__init__(name)

    def __call__(self, state=None, action=None, new_state=None, **kwargs) -> float:
        return np.random.random()


class CostFunc(RewardFunc):
    pass


class RewardFuncCostWrapper(CostFunc):

    def __init__(self, reward_func: RewardFunc, name='reward_func'):
        super().__init__(name)
        self._reward_func = reward_func

    def __call__(self, state, action, new_state, **kwargs) -> float:
        return self._reward_func.__call__(state=state,
                                          action=action,
                                          new_state=new_state) * -1.0

    def init(self):
        self._reward_func.init()


class QuadraticCostFunc(CostFunc):
    """
    A quadratic function
    """

    def __init__(self, C, c, name='reward_func'):
        """
        the cost is computed by 1/2[x_t, u_t].T *  C * [x_t, u_t] + [x_t, u_t] * c
        :param C: quadratic term
        :param c: linear term
        :param name:
        """
        super().__init__(name)
        self.C = np.array(C)
        self.c = np.array(c)
        self.state_action_flat_dim = self.C.shape[0]
        assert self.state_action_flat_dim == self.C.shape[1]
        assert len(self.C.shape) == 2
        assert self.c.shape[0] == self.state_action_flat_dim

    def __call__(self, state=None, action=None, new_state=None, **kwargs) -> float:
        u_s = np.concatenate((np.array(state).reshape(-1), np.array(action).reshape(-1))).reshape(
            self.state_action_flat_dim, 1)
        res = 0.5 * np.dot(np.dot(u_s.T, self.C), u_s) + np.dot(u_s.T, self.c).reshape(())
        return float(res)



import abc
import logging
import os

from baconian.common.misc import construct_dict_config
from baconian.common import files as files
from baconian.core.global_var import get_all
from baconian.config.global_config import GlobalConfig
from functools import wraps
from baconian.common.error import *


class BaseLogger(object):
    required_key_dict = ()

    def __init__(self):
        self.inited_flag = False

    @abc.abstractmethod
    def close(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def init(self, *args, **kwargs):
        raise NotImplementedError


class _SingletonConsoleLogger(BaseLogger):
    """
    A private class that should never be instanced, it is used to implement the singleton design pattern for
    ConsoleLogger
    """
    ALLOWED_LOG_LEVEL = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET')
    ALLOWED_PRINT_TYPE = ('info', 'warning', 'debug', 'critical', 'log', 'critical', 'error')

    def __init__(self):
        super(_SingletonConsoleLogger, self).__init__()
        self.name = None
        self.logger = None

    def init(self, to_file_flag, level: str, to_file_name: str = None, logger_name: str = 'console_logger'):
        if self.inited_flag is True:
            return
        self.name = logger_name
        if level not in self.ALLOWED_LOG_LEVEL:
            raise ValueError('Wrong log level use {} instead'.format(self.ALLOWED_LOG_LEVEL))
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(getattr(logging, level))

        for handler in self.logger.root.handlers[:] + self.logger.handlers[:]:
            self.logger.removeHandler(handler)
            self.logger.root.removeHandler(handler)

        self.logger.addHandler(logging.StreamHandler())
        if to_file_flag is True:
            self.logger.addHandler(logging.FileHandler(filename=to_file_name))
        for handler in self.logger.root.handlers[:] + self.logger.handlers[:]:
            handler.setFormatter(fmt=logging.Formatter(fmt=GlobalConfig().DEFAULT_LOGGING_FORMAT))
            handler.setLevel(getattr(logging, level))

        self.inited_flag = True

    def print(self, p_type: str, p_str: str, *arg, **kwargs):
        if p_type not in self.ALLOWED_PRINT_TYPE:
            raise ValueError('use print type from {}'.format(self.ALLOWED_PRINT_TYPE))
        getattr(self.logger, p_type)(p_str, *arg, **kwargs)
        self.flush()

    def close(self):
        self.flush()
        for handler in self.logger.root.handlers[:] + self.logger.handlers[:]:
            handler.close()
            self.logger.removeHandler(handler)
            self.logger.root.removeHandler(handler)

    def reset(self):
        self.close()
        self.inited_flag = False

    def flush(self):
        for handler in self.logger.root.handlers[:] + self.logger.handlers[:]:
            handler.flush()


class _SingletonLogger(BaseLogger):
    """
    A private class that should never be instanced, it is used to implement the singleton design pattern for Logger
    """

    def __init__(self):
        super(_SingletonLogger, self).__init__()
        self._registered_recorders = []
        self._log_dir = None
        self._config_file_log_dir = None
        self._record_file_log_dir = None
        self.logger_config = None
        self.log_level = None

    def init(self, config_or_config_dict,
             log_path, log_level=None, **kwargs):
        if self.inited_flag:
            return
        self._log_dir = log_path
        # if os.path.exists(self._log_dir):
        #     raise FileExistsError('%s path is existed' % self._log_dir)
        self._config_file_log_dir = os.path.join(self._log_dir, 'config')
        self._record_file_log_dir = os.path.join(self._log_dir, 'record')

        self.logger_config = construct_dict_config(config_or_config_dict, obj=self)
        self.log_level = log_level
        self.inited_flag = True

    @property
    def log_dir(self):
        if os.path.exists(self._log_dir) is False:
            os.makedirs(self._log_dir)
        return self._log_dir

    def flush_recorder(self, recorder=None):
        if not recorder:
            for re in self._registered_recorders:
                self._flush(re)
        else:
            self._flush(recorder)

    def close(self):
        self._save_all_obj_final_status()
        self.flush_recorder()
        self._registered_recorders = []

    def append_recorder(self, recorder):
        self._registered_recorders.append(recorder)

    def reset(self):
        self.close()
        for re in self._registered_recorders:
            re.reset()
        self._registered_recorders = []
        self.inited_flag = False

    def _flush(self, recorder):
        if recorder.is_empty():
            return
        log_dict, by_status_flag = recorder.get_obj_log_to_flush(clear_obj_log_flag=True)
        for obj_name, obj_log_dict in log_dict.items():
            if by_status_flag is True:
                for status, status_log_dict in obj_log_dict.items():
                    ConsoleLogger().print('info', 'save {}, with status: {} log into {}'.format(str(obj_name),
                                                                                                str(status),
                                                                                                os.path.join(
                                                                                                    self._record_file_log_dir,
                                                                                                    str(obj_name),
                                                                                                    str(status))))
                    self.out_to_file(
                        file_path=os.path.join(self._record_file_log_dir, str(obj_name), str(status)),
                        content=status_log_dict,
                        file_name='log.json')
            else:
                ConsoleLogger().print('info', 'save {} log into {}'.format(str(obj_name),
                                                                           os.path.join(
                                                                               self._record_file_log_dir,
                                                                               str(obj_name))))
                self.out_to_file(file_path=os.path.join(self._record_file_log_dir, str(obj_name)),
                                 content=obj_log_dict,
                                 file_name='log.json')

    def _save_all_obj_final_status(self):
        final_status = dict()
        for obj_name, obj in get_all()['_global_name_dict'].items():
            if hasattr(obj, 'get_status') and callable(getattr(obj, 'get_status')):
                tmp_dict = dict()
                tmp_dict[obj_name] = dict()
                for st in obj.STATUS_LIST:
                    obj.set_status(st)
                    tmp_dict[obj_name][st] = obj.get_status()
                final_status = {**final_status, **tmp_dict}
        ConsoleLogger().print('info', 'save final_status into {}'.format(os.path.join(
            self._record_file_log_dir)))
        self.out_to_file(file_path=os.path.join(self._record_file_log_dir),
                         content=final_status,
                         force_new=True,
                         file_name='final_status.json')
        ConsoleLogger().print('info', 'save global_config into {}'.format(os.path.join(
            self._record_file_log_dir)))
        self.out_to_file(file_path=os.path.join(self._record_file_log_dir),
                         content=GlobalConfig().return_all_as_dict(),
                         force_new=True,
                         file_name='global_config.json')

    @staticmethod
    def out_to_file(file_path: str, content: (tuple, list, dict), file_name: str, force_new=False):

        if len(content) == 0:
            return
        if force_new is True:
            mode = 'w'
        else:
            mode = 'a'
        if not os.path.exists(file_path):
            os.makedirs(file_path)
            mode = 'w'
        try:
            f = open(os.path.join(file_path, file_name), mode)
        except FileNotFoundError:
            f = open(os.path.join(file_path, file_name), 'w')
        files.save_to_json(content, fp=f)


class Logger(object):
    only_instance = None

    def __new__(cls, *args, **kwargs):
        if Logger.only_instance is None:
            Logger.only_instance = _SingletonLogger()
        return Logger.only_instance


class ConsoleLogger(object):
    only_instance = None

    def __new__(cls, *args, **kwargs):
        if not ConsoleLogger.only_instance:
            ConsoleLogger.only_instance = _SingletonConsoleLogger()
        return ConsoleLogger.only_instance


class Recorder(object):
    def __init__(self, flush_by_split_status=True, default_obj=None):
        self._obj_log = {}
        self._registered_log_attr_by_get_dict = {}
        Logger().append_recorder(self)
        self.flush_by_split_status = flush_by_split_status
        self._default_obj = default_obj

    def append_to_obj_log(self, obj, attr_name: str, status_info: dict, log_val):
        assert hasattr(obj, 'name')
        if obj not in self._obj_log:
            self._obj_log[obj] = {}
        if attr_name not in self._obj_log[obj]:
            self._obj_log[obj][attr_name] = []
        self._obj_log[obj][attr_name].append(dict(**status_info, attr_name=attr_name, log_val=log_val))

    def get_log(self, attr_name: str, filter_by_status: dict = None, obj=None):
        if obj is None:
            obj = self._default_obj
        if obj not in self._obj_log:
            raise LogItemNotExisted('object {} has no records in this recorder'.format(obj))
        if attr_name not in self._obj_log[obj]:
            raise LogItemNotExisted('no log item {} found at object {} recorder'.format(attr_name, obj))
        record = self._obj_log[obj][attr_name]
        if filter_by_status is not None:
            # TODO reduce the time complexity of the code snippet
            filtered_record = []
            for r in record:
                not_equal_flag = False
                for key in filter_by_status.keys():
                    if key in r and r[key] != filter_by_status[key]:
                        not_equal_flag = True
                if not not_equal_flag:
                    filtered_record.append(r)
            return filtered_record
        else:
            return record

    def is_empty(self):
        return len(self._obj_log) == 0

    def record(self):
        self._record_by_getter()

    def register_logging_attribute_by_record(self, obj, attr_name: str, static_flag: bool,
                                             get_method=None):
        """
        register an attribute that will be recorded periodically during training, duplicated registered will be ignored

        :param obj:
        :param attr_name:
        :param static_flag:
        :param get_method:
        :return:
        """
        if not hasattr(obj, 'get_status') or not callable(obj.get_status):
            raise ValueError('registered obj {} mush have callable method get_status()'.format(type(obj)))
        if obj not in self._registered_log_attr_by_get_dict:
            self._registered_log_attr_by_get_dict[obj] = {}
        if attr_name in self._registered_log_attr_by_get_dict[obj]:
            return
        self._registered_log_attr_by_get_dict[obj][attr_name] = dict(obj=obj,
                                                                     attr_name=attr_name,
                                                                     get_method=get_method,
                                                                     static_flag=static_flag)

    def _filter_by_main_status(self, clear_obj_log_flag=True):
        filtered_res = dict()
        for obj in self._obj_log:
            filtered_res[obj.name] = dict()
            status_list = obj.status_list
            for stat in status_list:
                filtered_res[obj.name][stat] = dict()
                for attr in self._obj_log[obj]:
                    filtered_res[obj.name][stat][attr] = []
            for attr in self._obj_log[obj]:
                for val_dict in self._obj_log[obj][attr]:
                    filtered_res[obj.name][val_dict['status']][val_dict['attr_name']].append(val_dict)
                    filtered_res[obj.name][val_dict['status']][attr][-1].pop('attr_name')
        if clear_obj_log_flag is True:
            del self._obj_log
            self._obj_log = {}
        return filtered_res

    def get_obj_log_to_flush(self, clear_obj_log_flag) -> (dict, bool):
        if self.flush_by_split_status is True:
            return self._filter_by_main_status(clear_obj_log_flag), self.flush_by_split_status
        else:
            filtered_res = {}
            for obj in self._obj_log:
                filtered_res[obj.name] = dict()
                for attr in self._obj_log[obj]:
                    filtered_res[obj.name][attr] = []
                    for val_dict in self._obj_log[obj][attr]:
                        val_dict.pop('attr_name')
                        filtered_res[obj.name][attr].append(val_dict)
            del self._obj_log
            self._obj_log = {}
            return filtered_res, self.flush_by_split_status

    def reset(self):
        self._obj_log = {}
        self._registered_log_attr_by_get_dict = {}

    def flush(self):
        Logger().flush_recorder(recorder=self)

    def _record_by_getter(self):
        for key, obj_dict in self._registered_log_attr_by_get_dict.items():
            for _, val in obj_dict.items():
                if val['get_method'] is None:
                    res = val['obj'].__getattribute__(val['attr_name'])
                else:
                    res = val['get_method'](val)
                self.append_to_obj_log(obj=val['obj'],
                                       attr_name=val['attr_name'],
                                       log_val=res,
                                       status_info=val['obj'].get_status())


def record_return_decorator(which_recorder: str = 'global'):
    def wrap(fn):
        @wraps(fn)
        def wrap_with_self(self, *args, **kwargs):
            obj = self
            if which_recorder == 'global':
                recorder = get_global_recorder()
            elif which_recorder == 'self':
                recorder = getattr(obj, 'recorder')
            else:
                raise ValueError('Not supported recorder indicator: {}, use {}'.format(which_recorder, 'gloabl, self'))
            if not hasattr(obj, 'get_status') or not callable(obj.get_status):
                raise ValueError('registered obj {} mush have callable method get_status()'.format(type(obj)))
            res = fn(self, *args, **kwargs)
            if res is not None:
                info = obj.get_status()
                if not isinstance(res, dict):
                    raise TypeError('returned value by {} must be a dict in order to be recorded'.format(fn.__name__))
                for key, val in res.items():
                    recorder.append_to_obj_log(obj=obj, attr_name=key, status_info=info,
                                               log_val=val)
            return res

        return wrap_with_self

    return wrap


_global_recorder = Recorder()


def get_global_recorder() -> Recorder:
    return globals()['_global_recorder']


def reset_global_recorder():
    globals()['_global_recorder'].reset()


def reset_logging():
    Logger().reset()
    ConsoleLogger().reset()
    reset_global_recorder()



from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import GlobalDynamicsModel

from baconian.core.parameters import Parameters
import numpy as np


class UniformRandomDynamicsModel(GlobalDynamicsModel):
    """
    A dynamics that uniformly return the new state (sample by env_spec.obs_space.sample()),
    can be used for debugging.
    """

    def __init__(self, env_spec: EnvSpec, parameters: Parameters = None, init_state=None, name='dynamics_model'):
        super().__init__(env_spec, parameters, init_state, name)

    def init(self):
        super().init()

    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        return self.env_spec.obs_space.sample()

    def make_copy(self):
        return UniformRandomDynamicsModel(env_spec=self.env_spec)



from baconian.core.core import Basic, Env
from baconian.common.sampler.sample_data import TransitionData, TrajectoryData
from typeguard import typechecked
import numpy as np


class Sampler(object):
    """
    Sampler module that handle the sampling procedure for training/testing of the agent.
    """

    @staticmethod
    @typechecked
    def sample(env: Env,
               agent,
               sample_count: int,
               sample_type='transition',
               reset_at_start=None) -> (TransitionData, TrajectoryData):
        """
        a static method of sample function

        :param env: environment object to sample from.
        :param agent: agent object to offer the sampling policy
        :param in_which_status: a string, "TEST" or "TRAIN" indicate this sample is used for training or testing
                                            (evaluation)

        :param sample_count: number of samples. If the sample_type == "transition", then this value means the number of
                                    transitions, usually for off-policy method like DQN, DDPG. If the sample_type ==
                                    "trajectory", then this value means the numbers of trajectories.

        :param sample_type: a string, "transition" or "trajectory".
        :param reset_at_start: A bool, if True, will reset the environment at the beginning, if False, continue sampling
                                    based on previous state (this is useful for certain tasks that you need to preserve
                                    previous state to reach the terminal goal state). If None, for sample_type ==
                                    "transition", it will set to False, for sample_type == "trajectory", it will set to True.

        :return: SampleData object.
        """
        state = None
        if reset_at_start is True or (reset_at_start is None and sample_type == 'trajectory'):
            state = env.reset()
        elif reset_at_start is False or (reset_at_start is None and sample_type == 'transition'):
            state = env.get_state()
        if sample_type == 'transition':
            return Sampler._sample_transitions(env, agent, sample_count, state)
        elif sample_type == 'trajectory':
            return Sampler._sample_trajectories(env, agent, sample_count, state)
        else:
            raise ValueError()

    @staticmethod
    def _sample_transitions(env: Env, agent, sample_count, init_state):
        state = init_state
        sample_record = TransitionData(env_spec=env.env_spec)

        for i in range(sample_count):
            action = agent.predict(obs=state)
            new_state, re, done, info = env.step(action)
            if not isinstance(done, bool):
                raise TypeError()
            sample_record.append(state=state,
                                 action=action,
                                 reward=re,
                                 new_state=new_state,
                                 done=done)
            if done:
                state = env.reset()
                agent.reset_on_terminal_state()
            else:
                state = new_state
        return sample_record

    @staticmethod
    def _sample_trajectories(env, agent, sample_count, init_state):
        state = init_state
        sample_record = TrajectoryData(env.env_spec)
        done = False
        for i in range(sample_count):
            traj_record = TransitionData(env.env_spec)
            while done is not True:
                action = agent.predict(obs=state)
                new_state, re, done, info = env.step(action)
                if not isinstance(done, bool):
                    raise TypeError()
                traj_record.append(state=state,
                                   action=action,
                                   reward=re,
                                   new_state=new_state,
                                   done=done)
                state = new_state
            agent.reset_on_terminal_state()
            done = False
            state = env.reset()
            sample_record.append(traj_record)
        return sample_record



import numpy as np
from typeguard import typechecked
from baconian.common.sampler.sample_data import TransitionData, TrajectoryData, SampleData
from baconian.common.error import *


class RingBuffer(object):
    @typechecked
    def __init__(self, maxlen: int, shape: (list, tuple), dtype='float32'):
        self.maxlen = maxlen
        self.start = 0
        self.length = 0
        self.data = np.zeros((maxlen,) + shape).astype(dtype)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        if idx < 0 or idx >= self.length:
            raise KeyError()
        return self.data[(self.start + idx) % self.maxlen]

    def get_batch(self, idxs):
        return self.data[(self.start + idxs) % self.maxlen]

    def append(self, v):
        if self.length < self.maxlen:
            # We have space, simply increase the length.
            self.length += 1
        elif self.length == self.maxlen:
            # No space, "remove" the first item.
            self.start = (self.start + 1) % self.maxlen
        else:
            # This should never happen.
            raise RuntimeError()
        self.data[(self.start + self.length - 1) % self.maxlen] = v


def array_min2d(x):
    x = np.array(x)

    if x.ndim >= 2:
        return x
    return x.reshape(-1, 1)


class BaseReplayBuffer(object):
    def __init__(self, limit, action_shape, observation_shape):
        self.limit = limit
        self.action_shape = action_shape
        self.obs_shape = observation_shape

        self.observations0 = RingBuffer(limit, shape=observation_shape)
        self.actions = RingBuffer(limit, shape=action_shape)
        self.rewards = RingBuffer(limit, shape=(1,))
        self.terminals1 = RingBuffer(limit, shape=(1,))
        self.observations1 = RingBuffer(limit, shape=observation_shape)

    def sample(self, batch_size):
        raise NotImplementedError

    def append(self, obs0, obs1, action, reward, terminal1, training=True):
        if not training:
            return
        self.observations0.append(obs0)
        self.actions.append(action)
        self.rewards.append(reward)
        self.observations1.append(obs1)
        self.terminals1.append(terminal1)

    @property
    def nb_entries(self):
        return len(self.observations0)

    def reset(self):
        self.observations0 = RingBuffer(self.limit, shape=self.obs_shape)
        self.actions = RingBuffer(self.limit, shape=self.action_shape)
        self.rewards = RingBuffer(self.limit, shape=(1,))
        self.terminals1 = RingBuffer(self.limit, shape=(1,))
        self.observations1 = RingBuffer(self.limit, shape=self.obs_shape)


class UniformRandomReplayBuffer(BaseReplayBuffer):
    def __init__(self, limit, action_shape, observation_shape):
        super().__init__(limit, action_shape, observation_shape)

    def sample(self, batch_size) -> SampleData:
        if self.nb_entries < batch_size:
            raise MemoryBufferLessThanBatchSizeError()

        batch_idxs = np.random.randint(self.nb_entries - 2, size=batch_size)
        pass

        obs0_batch = self.observations0.get_batch(batch_idxs)
        obs1_batch = self.observations1.get_batch(batch_idxs)
        action_batch = self.actions.get_batch(batch_idxs)
        reward_batch = self.rewards.get_batch(batch_idxs)
        terminal1_batch = self.terminals1.get_batch(batch_idxs)

        result = {
            'obs0': array_min2d(obs0_batch),
            'obs1': array_min2d(obs1_batch),
            'rewards': array_min2d(reward_batch),
            'actions': array_min2d(action_batch),
            'terminals1': array_min2d(terminal1_batch),
        }

        res = TransitionData(obs_shape=self.obs_shape, action_shape=self.action_shape)
        for obs0, obs1, action, terminal, re in zip(result['obs0'], result['obs1'], result['actions'],
                                                    result['terminals1'], result['rewards']):
            res.append(state=obs0, new_state=obs1, action=action, done=terminal, reward=re)
        return res
        pass


class PrioritisedReplayBuffer(BaseReplayBuffer):
    def __init__(self, limit, action_shape, observation_shape, alpha, beta, beta_increment):
        super().__init__(limit, action_shape, observation_shape)

        it_capacity = 1
        while it_capacity < limit:
            it_capacity *= 2

        assert alpha >= 0
        self.alpha = alpha
        self.beta = beta
        self.beta_increment = beta_increment
        self.max_priority = 1.0

    def update_priorities(self, idxes, priorities):
        assert len(idxes) == len(priorities)
        for idx, priority in zip(idxes, priorities):
            assert priority > 0
            assert 0 <= idx < len(self.observations0)
            self.it_sum[idx] = priority ** self.alpha
            self.it_min[idx] = priority ** self.alpha

            self.max_priority = max(self.max_priority, priority)

    def sample(self, batch_size) -> SampleData:
        if self.nb_entries < batch_size:
            raise MemoryBufferLessThanBatchSizeError()

        # todo This will be changed to prioritised
        batch_idxs = np.random.randint(self.nb_entries - 2, size=batch_size)

        obs0_batch = self.observations0.get_batch(batch_idxs)
        obs1_batch = self.observations1.get_batch(batch_idxs)
        action_batch = self.actions.get_batch(batch_idxs)
        reward_batch = self.rewards.get_batch(batch_idxs)
        terminal1_batch = self.terminals1.get_batch(batch_idxs)

        result = {
            'obs0': array_min2d(obs0_batch),
            'obs1': array_min2d(obs1_batch),
            'rewards': array_min2d(reward_batch),
            'actions': array_min2d(action_batch),
            'terminals1': array_min2d(terminal1_batch),
        }

        res = TransitionData(obs_shape=self.obs_shape, action_shape=self.action_shape)
        for obs0, obs1, action, terminal, re in zip(result['obs0'], result['obs1'], result['actions'],
                                                    result['terminals1'], result['rewards']):
            res.append(state=obs0, new_state=obs1, action=action, done=terminal, reward=re)
        return res


from baconian.common.log_data_loader import *

exp_root_dir_list = [
    '/Users/lukeeeeee/Code/baconian-internal/examples/log/',
]

if __name__ == "__main__":
    MultipleExpLogDataLoader(exp_root_dir_list=exp_root_dir_list) \
        .plot_res(sub_log_dir_name='demo_exp_agent/TRAIN',
                  key='sum_reward',
                  index='sample_counter',
                  mode='line',
                  average_over=1,
                  file_name=None,
                  save_format='png',
                  save_path='/baconian/examples/log')



"""
Module that compute diagonal multivariate normal distribution operation with tensorflow tensor as parameters
"""
import tensorflow as tf
import numpy as np


def kl(mean_p, var_p, mean_q, var_q, dims):
    """
    Compute the KL divergence of diagonal multivariate normal distribution q, and p, which is KL(P||Q)
    :param mean_p:
    :param var_p:
    :param mean_q:
    :param var_q:
    :param dims:
    :return:
    """
    # p is old

    log_var_p = tf.log(var_p)
    log_var_q = tf.log(var_q)
    log_det_cov_p = tf.reduce_sum(log_var_p)
    log_det_cov_q = tf.reduce_sum(log_var_q)
    tr_p_q = tf.reduce_sum(tf.exp(log_var_p - log_var_q))
    kl = 0.5 * tf.reduce_mean(
        log_det_cov_q - log_det_cov_p + tr_p_q + tf.reduce_sum(tf.square(mean_q - mean_p) / tf.exp(log_det_cov_q),
                                                               axis=1) - dims)
    return kl


def entropy(mean_p, var_p, dims):
    return 0.5 * (dims * (np.log(2 * np.pi) + 1) + tf.reduce_sum(tf.log(var_p)))


def log_prob(variable_ph, mean_p, var_p):
    log_prob = -0.5 * (tf.reduce_sum(tf.log(var_p)) + (np.log(2 * np.pi)))
    log_prob += -0.5 * tf.reduce_sum(tf.square(variable_ph - mean_p) / var_p, axis=1)
    return log_prob


def prob(variable_ph, mean_p, var_p):
    return tf.log(log_prob(variable_ph, mean_p, var_p))


def sample(mean_p, var_p, dims):
    return mean_p + tf.math.sqrt(var_p) * tf.random_normal(shape=(dims,))



"""
A simple example to show how to build up an experiment with Dyna training and testing on Pendulum-v0
"""
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.algo.ddpg import DDPG
from baconian.algo.policy import DeterministicMLPPolicy
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.config.global_config import GlobalConfig
from baconian.core.status import get_global_status_collect
from baconian.common.schedules import PeriodicalEventSchedule
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.algo.dyna import Dyna
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc
from baconian.algo.dynamics.terminal_func.terminal_func import FixedEpisodeLengthTerminalFunc
from baconian.core.flow.dyna_flow import create_dyna_flow
from baconian.common.data_pre_processing import RunningStandardScaler


def task_fn():
    # create the gym environment by make function
    env = make('Pendulum-v0')
    # give your experiment a name which is used to generate the log path etc.
    name = 'demo_exp'
    # construct the environment specification
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)
    # construct the neural network to approximate q function of DDPG
    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    # construct the neural network to approximate policy for DDPG
    policy = DeterministicMLPPolicy(env_spec=env_spec,
                                    name_scope=name + '_mlp_policy',
                                    name=name + '_mlp_policy',
                                    mlp_config=[
                                        {
                                            "ACT": "RELU",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "1",
                                            "N_UNITS": 16,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        },
                                        {
                                            "ACT": "LINEAR",
                                            "B_INIT_VALUE": 0.0,
                                            "NAME": "OUPTUT",
                                            "N_UNITS": env_spec.flat_action_dim,
                                            "TYPE": "DENSE",
                                            "W_NORMAL_STDDEV": 0.03
                                        }
                                    ],
                                    reuse=False)
    # construct the DDPG algorithms
    ddpg = DDPG(
        env_spec=env_spec,
        config_or_config_dict={
            "REPLAY_BUFFER_SIZE": 10000,
            "GAMMA": 0.999,
            "CRITIC_LEARNING_RATE": 0.001,
            "ACTOR_LEARNING_RATE": 0.001,
            "DECAY": 0.5,
            "BATCH_SIZE": 50,
            "TRAIN_ITERATION": 1,
            "critic_clip_norm": 0.1,
            "actor_clip_norm": 0.1,
        },
        value_func=mlp_q,
        policy=policy,
        name=name + '_ddpg',
        replay_buffer=None
    )
    # construct a neural network based global dynamics model to approximate the state transition of environment
    mlp_dyna = ContinuousMLPGlobalDynamicsModel(
        env_spec=env_spec,
        name_scope=name + '_mlp_dyna',
        name=name + '_mlp_dyna',
        learning_rate=0.01,
        state_input_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
        action_input_scaler=RunningStandardScaler(dims=env_spec.flat_action_dim),
        output_delta_state_scaler=RunningStandardScaler(dims=env_spec.flat_obs_dim),
        mlp_config=[
            {
                "ACT": "RELU",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "L1_NORM": 0.0,
                "L2_NORM": 0.0,
                "N_UNITS": 16,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "LINEAR",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "L1_NORM": 0.0,
                "L2_NORM": 0.0,
                "N_UNITS": env_spec.flat_obs_dim,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ])
    # finally, construct the Dyna algorithms with a model free algorithm DDGP, and a NN model.
    algo = Dyna(env_spec=env_spec,
                name=name + '_dyna_algo',
                model_free_algo=ddpg,
                dynamics_model=mlp_dyna,
                config_or_config_dict=dict(
                    dynamics_model_train_iter=10,
                    model_free_algo_train_iter=10
                ))
    # To make the NN based dynamics model a proper environment so be a sampling source for DDPG, reward function and
    # terminal function need to be set.

    # For examples only, we use random reward function and terminal function with fixed episode length.
    algo.set_terminal_reward_function_for_dynamics_env(
        terminal_func=FixedEpisodeLengthTerminalFunc(max_step_length=env.unwrapped._max_episode_steps,
                                                     step_count_fn=algo.dynamics_env.total_step_count_fn),
        reward_func=RandomRewardFunc())
    # construct agent with additional exploration strategy if needed.
    agent = Agent(env=env, env_spec=env_spec,
                  algo=algo,
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=20,
                      after_t=10),
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))
    # construct the training flow, called Dyna flow. It defines how the training proceed, and the terminal condition
    flow = create_dyna_flow(
        train_algo_func=(agent.train, (), dict(state='state_agent_training')),
        train_algo_from_synthesized_data_func=(agent.train, (), dict(state='state_agent_training')),
        train_dynamics_func=(agent.train, (), dict(state='state_dynamics_training')),
        test_algo_func=(agent.test, (), dict(sample_count=1)),
        test_dynamics_func=(agent.algo.test_dynamics, (), dict(sample_count=10, env=env)),
        sample_from_real_env_func=(agent.sample, (), dict(sample_count=10,
                                                          env=agent.env,
                                                          store_flag=True)),
        sample_from_dynamics_env_func=(agent.sample, (), dict(sample_count=10,
                                                              env=agent.algo.dynamics_env,
                                                              store_flag=True)),
        train_algo_every_real_sample_count_by_data_from_real_env=40,
        train_algo_every_real_sample_count_by_data_from_dynamics_env=40,
        test_algo_every_real_sample_count=40,
        test_dynamics_every_real_sample_count=40,
        train_dynamics_ever_real_sample_count=20,
        start_train_algo_after_sample_count=1,
        start_train_dynamics_after_sample_count=1,
        start_test_algo_after_sample_count=1,
        start_test_dynamics_after_sample_count=1,
        warm_up_dynamics_samples=1
    )
    # construct the experiment
    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name + '_exp'
    )
    # run!
    experiment.run()


from baconian.core.experiment_runner import *

# set some global configuration here

# set DEFAULT_EXPERIMENT_END_POINT to indicate when to stop the experiment.
# one usually used is the TOTAL_AGENT_TRAIN_SAMPLE_COUNT, i.e., how many samples/timesteps are used for training
GlobalConfig().set('DEFAULT_EXPERIMENT_END_POINT', dict(TOTAL_AGENT_TRAIN_SAMPLE_COUNT=200))

# set the logging path to write log and save model checkpoints.
GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')

# feed the task into a exp runner.
single_exp_runner(task_fn, del_if_log_path_existed=True)



from baconian.common.data_pre_processing import DataScaler
from baconian.core.core import EnvSpec
from baconian.algo.dynamics.dynamics_model import GlobalDynamicsModel, TrainableDyanmicsModel

from baconian.core.parameters import Parameters
import numpy as np
from copy import deepcopy
from sklearn.linear_model import LinearRegression
from baconian.common.sampler.sample_data import TransitionData


class LinearDynamicsModel(GlobalDynamicsModel):
    """
    A linear dynamics model given the transition matrix F and the bias f (can't be trained, use LinearRegressionModel instead if your want to fit one)
    """

    def __init__(self, env_spec: EnvSpec, state_transition_matrix: np.array, bias: np.array, init_state=None,
                 name='dynamics_model', state_input_scaler: DataScaler = None, action_input_scaler: DataScaler = None,
                 state_output_scaler: DataScaler = None):
        parameters = Parameters(parameters=dict(F=state_transition_matrix, f=bias))
        super().__init__(env_spec, parameters, init_state, name, state_input_scaler, action_input_scaler,
                         state_output_scaler)

        assert self.parameters('F').shape == \
               (env_spec.obs_space.flat_dim, env_spec.obs_space.flat_dim + env_spec.action_space.flat_dim)
        assert self.parameters('f').shape[0] == env_spec.obs_space.flat_dim

    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        new_state = np.dot(self.parameters('F'), np.concatenate((state, action))) + self.parameters('f')
        return self.env_spec.obs_space.clip(new_state)

    def make_copy(self):
        return LinearDynamicsModel(env_spec=self.env_spec,
                                   state_transition_matrix=deepcopy(self.parameters('F')),
                                   bias=deepcopy(self.parameters('f')))

    @property
    def F(self):
        return self.parameters('F')

    @property
    def f(self):
        return self.parameters('f')


class LinearRegressionDynamicsModel(GlobalDynamicsModel, TrainableDyanmicsModel):

    def __init__(self, env_spec: EnvSpec, init_state=None, name='dynamics_model',
                 state_input_scaler: DataScaler = None, action_input_scaler: DataScaler = None,
                 state_output_scaler: DataScaler = None):
        super().__init__(env_spec=env_spec, init_state=init_state, name=name,
                         state_input_scaler=state_input_scaler,
                         action_input_scaler=action_input_scaler,
                         state_output_scaler=state_output_scaler)
        self._linear_model = LinearRegression(fit_intercept=True,
                                              normalize=False)

    def _state_transit(self, state, action, **kwargs) -> np.ndarray:
        state = self.state_input_scaler.process(np.array(state).reshape(self.env_spec.obs_shape))
        action = self.action_input_scaler.process(action.reshape(self.env_spec.flat_action_dim))
        new_state = self._linear_model.predict(np.concatenate([state, action], axis=-1).reshape(1, -1))
        new_state = np.clip(self.state_output_scaler.inverse_process(new_state),
                            self.env_spec.obs_space.low,
                            self.env_spec.obs_space.high).squeeze()
        return new_state

    def train(self, batch_data: TransitionData = None, *kwargs):
        self.state_input_scaler.update_scaler(batch_data.state_set)
        self.action_input_scaler.update_scaler(batch_data.action_set)
        self.state_output_scaler.update_scaler(batch_data.new_state_set)

        state = self.state_input_scaler.process(batch_data.state_set)
        action = self.action_input_scaler.process(batch_data.action_set)
        new_state = self.state_input_scaler.process(batch_data.new_state_set)

        self._linear_model.fit(X=np.concatenate([state, action], axis=-1),
                               y=new_state)



"""
A simple example to show how to build up an experiment with ddpg training and testing on MountainCarContinuous-v0
"""
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import create_train_test_flow
from baconian.algo.mpc import ModelPredictiveControl
from baconian.algo.dynamics.terminal_func.terminal_func import RandomTerminalFunc
from baconian.algo.dynamics.reward_func.reward_func import RandomRewardFunc
from baconian.algo.policy import UniformRandomPolicy
from baconian.algo.dynamics.mlp_dynamics_model import ContinuousMLPGlobalDynamicsModel
from baconian.config.global_config import GlobalConfig


def task_fn():
    env = make('Pendulum-v0')
    name = 'demo_exp'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_dyna = ContinuousMLPGlobalDynamicsModel(
        env_spec=env_spec,
        name_scope=name + '_mlp_dyna',
        name=name + '_mlp_dyna',
        learning_rate=0.01,
        mlp_config=[
            {
                "ACT": "RELU",
                "B_INIT_VALUE": 0.0,
                "NAME": "1",
                "L1_NORM": 0.0,
                "L2_NORM": 0.0,
                "N_UNITS": 16,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            },
            {
                "ACT": "LINEAR",
                "B_INIT_VALUE": 0.0,
                "NAME": "OUPTUT",
                "L1_NORM": 0.0,
                "L2_NORM": 0.0,
                "N_UNITS": env_spec.flat_obs_dim,
                "TYPE": "DENSE",
                "W_NORMAL_STDDEV": 0.03
            }
        ])
    algo = ModelPredictiveControl(
        dynamics_model=mlp_dyna,
        env_spec=env_spec,
        config_or_config_dict=dict(
            SAMPLED_HORIZON=2,
            SAMPLED_PATH_NUM=5,
            dynamics_model_train_iter=10
        ),
        name=name + '_mpc',

        policy=UniformRandomPolicy(env_spec=env_spec, name='uni_policy')
    )
    algo.set_terminal_reward_function_for_dynamics_env(reward_func=RandomRewardFunc(name='reward_func'),
                                                       terminal_func=RandomTerminalFunc(name='random_terminal'), )
    agent = Agent(env=env, env_spec=env_spec,
                  algo=algo,
                  name=name + '_agent',
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     init_random_prob=0.5))
    flow = create_train_test_flow(
        test_every_sample_count=10,
        train_every_sample_count=10,
        start_test_after_sample_count=5,
        start_train_after_sample_count=5,
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=10)),
        sample_func_and_args=(agent.sample, (), dict(sample_count=100,
                                                     env=agent.env,
                                                     store_flag=True))
    )
    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name
    )
    experiment.run()


from baconian.core.experiment_runner import single_exp_runner

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
single_exp_runner(task_fn, del_if_log_path_existed=True)



from baconian.common.special import *
from baconian.core.core import EnvSpec
from copy import deepcopy
import typeguard as tg
from baconian.common.error import *


class SampleData(object):
    def __init__(self, env_spec: EnvSpec = None, obs_shape=None, action_shape=None):
        if env_spec is None and (obs_shape is None or action_shape is None):
            raise ValueError('At least env_spec or (obs_shape, action_shape) should be passed in')
        self.env_spec = env_spec
        self.obs_shape = env_spec.obs_shape if env_spec else obs_shape
        self.action_shape = env_spec.action_shape if env_spec else action_shape

    def reset(self):
        raise NotImplementedError

    def append(self, *args, **kwargs):
        raise NotImplementedError

    def union(self, sample_data):
        raise NotImplementedError

    def get_copy(self):
        raise NotImplementedError

    def __call__(self, set_name, **kwargs):
        raise NotImplementedError

    def append_new_set(self, name, data_set: (list, np.ndarray), shape: (tuple, list)):
        raise NotImplementedError

    def sample_batch(self, *args, **kwargs):
        raise NotImplementedError

    def return_generator(self, *args, **kwargs):
        raise NotImplementedError

    def apply_transformation(self, set_name, func, *args, **kwargs):
        raise NotImplementedError

    def apply_op(self, set_name, func, *args, **kwargs):
        raise NotImplementedError


class TransitionData(SampleData):
    def __init__(self, env_spec: EnvSpec = None, obs_shape=None, action_shape=None):
        super(TransitionData, self).__init__(env_spec=env_spec, obs_shape=obs_shape, action_shape=action_shape)

        self.cumulative_reward = 0.0
        self.step_count_per_episode = 0
        assert isinstance(self.obs_shape, (list, tuple))
        assert isinstance(self.action_shape, (list, tuple))
        self.obs_shape = list(self.obs_shape)
        self.action_shape = list(self.action_shape)

        self._internal_data_dict = {
            'state_set': [np.empty([0] + self.obs_shape), self.obs_shape],
            'new_state_set': [np.empty([0] + self.obs_shape), self.obs_shape],
            'action_set': [np.empty([0] + self.action_shape), self.action_shape],
            'reward_set': [np.empty([0]), []],
            'done_set': [np.empty([0], dtype=bool), []]
        }
        self.current_index = 0

    def __len__(self):
        return len(self._internal_data_dict['state_set'][0])

    def __call__(self, set_name, **kwargs):
        if set_name not in self._allowed_data_set_keys:
            raise ValueError('pass in set_name within {} '.format(self._allowed_data_set_keys))
        return make_batch(self._internal_data_dict[set_name][0],
                          original_shape=self._internal_data_dict[set_name][1])

    def reset(self):
        for key, data_set in self._internal_data_dict.items():
            self._internal_data_dict[key][0] = np.empty([0, *self._internal_data_dict[key][1]])
        self.cumulative_reward = 0.0
        self.step_count_per_episode = 0

    def append(self, state: np.ndarray, action: np.ndarray, new_state: np.ndarray, done: bool, reward: float):
        self._internal_data_dict['state_set'][0] = np.concatenate(
            (self._internal_data_dict['state_set'][0], np.reshape(state, [1] + self.obs_shape)), axis=0)
        self._internal_data_dict['new_state_set'][0] = np.concatenate(
            (self._internal_data_dict['new_state_set'][0], np.reshape(new_state, [1] + self.obs_shape)), axis=0)
        self._internal_data_dict['reward_set'][0] = np.concatenate(
            (self._internal_data_dict['reward_set'][0], np.reshape(reward, [1])), axis=0)
        self._internal_data_dict['done_set'][0] = np.concatenate(
            (self._internal_data_dict['done_set'][0], np.reshape(np.array(done, dtype=bool), [1])), axis=0)
        self._internal_data_dict['action_set'][0] = np.concatenate(
            (self._internal_data_dict['action_set'][0], np.reshape(action, [1] + self.action_shape)), axis=0)
        self.cumulative_reward += reward

    def union(self, sample_data):
        assert isinstance(sample_data, type(self))
        self.cumulative_reward += sample_data.cumulative_reward
        self.step_count_per_episode += sample_data.step_count_per_episode
        for key, val in self._internal_data_dict.items():
            assert self._internal_data_dict[key][1] == sample_data._internal_data_dict[key][1]
            self._internal_data_dict[key][0] = np.concatenate(
                (self._internal_data_dict[key][0], sample_data._internal_data_dict[key][0]), axis=0)

    def get_copy(self):
        obj = TransitionData(env_spec=self.env_spec, obs_shape=self.obs_shape, action_shape=self.action_shape)
        for key in self._internal_data_dict:
            obj._internal_data_dict[key] = deepcopy(self._internal_data_dict[key])
        return obj

    def append_new_set(self, name, data_set: (list, np.ndarray), shape: (tuple, list)):
        assert len(data_set) == len(self)
        assert len(np.array(data_set).shape) - 1 == len(shape)
        if len(shape) > 0:
            assert np.equal(np.array(data_set).shape[1:], shape).all()
        shape = tuple(shape)
        self._internal_data_dict[name] = [np.array(data_set), shape]

    def sample_batch(self, batch_size, shuffle_flag=True, **kwargs) -> dict:
        if shuffle_flag is False:
            raise NotImplementedError
        total_num = len(self)
        id_index = np.random.randint(low=0, high=total_num, size=batch_size)
        batch_data = dict()
        for key in self._internal_data_dict.keys():
            batch_data[key] = self(key)[id_index]
        return batch_data

    def get_mean_of(self, set_name):
        return self.apply_op(set_name=set_name, func=np.mean)

    def get_sum_of(self, set_name):
        return self.apply_op(set_name=set_name, func=np.sum)

    def apply_transformation(self, set_name, func, direct_apply=False, **func_kwargs):
        data = make_batch(self._internal_data_dict[set_name][0],
                          original_shape=self._internal_data_dict[set_name][1])
        transformed_data = make_batch(func(data, **func_kwargs),
                                      original_shape=self._internal_data_dict[set_name][1])
        if transformed_data.shape != data.shape:
            raise TransformationResultedToDifferentShapeError()
        elif direct_apply is True:
            self._internal_data_dict[set_name][0] = transformed_data
        return transformed_data

    def apply_op(self, set_name, func, **func_kwargs):
        data = make_batch(self._internal_data_dict[set_name][0],
                          original_shape=self._internal_data_dict[set_name][1])
        applied_op_data = np.array(func(data, **func_kwargs))
        return applied_op_data

    def shuffle(self, index: list = None):
        if not index:
            index = np.arange(len(self._internal_data_dict['state_set'][0]))
            np.random.shuffle(index)
        for key in self._internal_data_dict.keys():
            self._internal_data_dict[key][0] = self._internal_data_dict[key][0][index]

    def return_generator(self, batch_size=None, shuffle_flag=False, assigned_keys=None, infinite_run=False):
        if assigned_keys is None:
            assigned_keys = ('state_set', 'new_state_set', 'action_set', 'reward_set', 'done_set')
        # todo unit test should be tested, the dataset should not be shuffled, only the generator should be shuffled
        if shuffle_flag is True:
            self.shuffle()
        if batch_size is not None:
            if batch_size <= 0:
                raise ValueError()
            start = 0
            if infinite_run is True:
                while True:
                    end = min(start + batch_size, len(self))
                    yield deepcopy(
                        [make_batch(self._internal_data_dict[key][0][start: end], self._internal_data_dict[key][1])
                         for key in assigned_keys])
                    start = end % len(self)
            else:
                while start < len(self):
                    end = min(start + batch_size, len(self))
                    yield deepcopy(
                        [make_batch(self._internal_data_dict[key][0][start: end], self._internal_data_dict[key][1])
                         for key in assigned_keys])
                    start = end
        else:
            start = 0
            if infinite_run is True:
                while True:
                    yield deepcopy([self._internal_data_dict[key][0][start] for key in assigned_keys])
                    start = (start + 1) % len(self)
            else:
                for i in range(len(self)):
                    yield deepcopy([self._internal_data_dict[key][0][i] for key in assigned_keys])

    @property
    def _allowed_data_set_keys(self):
        return list(self._internal_data_dict.keys())

    @property
    def state_set(self):
        return self('state_set')

    @property
    def new_state_set(self):
        return self('new_state_set')

    @property
    def action_set(self):
        return self('action_set')

    @property
    def reward_set(self):
        return self('reward_set')

    @property
    def done_set(self):
        return self('done_set')


class TrajectoryData(SampleData):
    def __init__(self, env_spec=None, obs_shape=None, action_shape=None):
        super(TrajectoryData, self).__init__(env_spec=env_spec, obs_shape=obs_shape, action_shape=action_shape)
        self.trajectories = []

    def reset(self):
        self.trajectories = []

    def append(self, transition_data: TransitionData):
        self.trajectories.append(transition_data)

    def union(self, sample_data):
        if not isinstance(sample_data, type(self)):
            raise TypeError()
        self.trajectories += sample_data.trajectories

    def return_as_transition_data(self, shuffle_flag=False) -> TransitionData:
        transition_set = self.trajectories[0].get_copy()
        for i in range(1, len(self.trajectories)):
            transition_set.union(self.trajectories[i])
        if shuffle_flag is True:
            transition_set.shuffle()
        return transition_set

    def get_mean_of(self, set_name):
        tran = self.return_as_transition_data()
        return tran.get_mean_of(set_name)

    def get_sum_of(self, set_name):
        tran = self.return_as_transition_data()
        return tran.get_sum_of(set_name)

    def __len__(self):
        return len(self.trajectories)

    def get_copy(self):
        tmp_traj = TrajectoryData(env_spec=self.env_spec, obs_shape=self.obs_shape, action_shape=self.action_shape)
        for traj in self.trajectories:
            tmp_traj.append(transition_data=traj.get_copy())
        return tmp_traj

    def return_generator(self, batch_size=None, shuffle_flag=False):
        return self.return_as_transition_data(shuffle_flag=False).return_generator(batch_size=batch_size,
                                                                                   shuffle_flag=shuffle_flag)

    def apply_transformation(self, set_name, func, direct_apply=False, **func_kwargs):
        # TODO unit test
        for traj in self.trajectories:
            traj.apply_transformation(set_name, func, direct_apply, **func_kwargs)

    def apply_op(self, set_name, func, **func_kwargs):
        # TODO unit test
        res = []
        for traj in self.trajectories:
            res.append(traj.apply_op(set_name, func, **func_kwargs))
        return np.array(res)



class Space:
    """
    Provides a classification state spaces and action spaces,
    so you can write generic code that applies to any Environment.
    E.g. to choose a random action.
    """

    def sample(self, seed=0):
        """
        Uniformly randomly sample a random element of this space
        """
        raise NotImplementedError

    def contains(self, x):
        """
        Return boolean specifying if x is a valid
        member of this space
        """
        raise NotImplementedError

    def flatten(self, x):
        raise NotImplementedError

    def unflatten(self, x):
        raise NotImplementedError

    def flatten_n(self, xs):
        raise NotImplementedError

    def unflatten_n(self, xs):
        raise NotImplementedError

    @property
    def flat_dim(self):
        """
        The dimension of the flattened vector of the tensor representation
        """
        raise NotImplementedError

    def new_tensor_variable(self, name, extra_dims):
        """
        Create a tensor variable given the name and extra dimensions
        prepended
        :param name: name of the variable
        :param extra_dims: extra dimensions in the front
        :return: the created tensor variable
        """
        raise NotImplementedError

    def clip(self, x):
        raise NotImplementedError

    def bound(self):
        raise NotImplementedError



# import numpy as np
# import time
# import typeguard as tg
# import functools
#
#
# def random_snapshot(func):
#     @functools.wraps(func)
#     def wrapper(random_instance):
#         random_instance.state_snapshot = random_instance._np_random.get_state()
#         return func(random_instance)
#
#     return wrapper
#
#
# class Random(object):
#     """
#
#     A random utility that based on Numpy random module in order to better control the randomness of the system.
#     The python random is not recommended to use
#
#     """
#
#     @tg.typechecked
#     def __init__(self, seed: int = int(round(time.time() * 1000)) % (2 ** 32 - 1), global_state: bool = True):
#         self.seed = seed
#         if global_state:
#             self._np_random = np.random
#         else:
#             self._np_random = np.random.RandomState()
#         self._np_random.seed(seed)
#         self.state_snapshot = self._np_random.get_state()
#
#     @random_snapshot
#     def unwrapped(self):
#         return self._np_random
#
#     @tg.typechecked
#     def set_seed(self, seed: int):
#         self.seed = seed
#         self._np_random.seed(seed)
#
#     @tg.typechecked
#     def set_state(self, state: tuple):
#         self._np_random.set_state(state)
#
#     def reset_last_state(self):
#         self.set_state(self.state_snapshot)
#
#     def _register_all_np_method(self):
#         raise NotImplementedError
#
#
# if __name__ == '__main__':
#     r = Random()
#     print(r.unwrapped().rand(1, 1))
#     r.reset_last_state()
#     print(r.unwrapped().rand(1, 1))



from baconian.algo.policy.policy import DeterministicPolicy
from baconian.core.core import EnvSpec
import overrides
import tensorflow as tf
from baconian.tf.mlp import MLP
from baconian.tf.tf_parameters import ParametersWithTensorflowVariable
from baconian.common.special import *
from baconian.algo.utils import _get_copy_arg_with_tf_reuse
from baconian.algo.misc.placeholder_input import PlaceholderInput


class DeterministicMLPPolicy(DeterministicPolicy, PlaceholderInput):

    def __init__(self, env_spec: EnvSpec,
                 name_scope: str,
                 name: str, mlp_config: list,
                 input_norm: np.ndarray = None,
                 output_norm: np.ndarray = None,
                 output_low: np.ndarray = None,
                 output_high: np.ndarray = None,
                 reuse=False):
        DeterministicPolicy.__init__(self, env_spec=env_spec, name=name, parameters=None)
        obs_dim = env_spec.flat_obs_dim
        action_dim = env_spec.flat_action_dim
        assert action_dim == mlp_config[-1]['N_UNITS']

        with tf.variable_scope(name_scope):
            state_input = tf.placeholder(shape=[None, obs_dim], dtype=tf.float32, name='state_ph')

        mlp_kwargs = dict(
            reuse=reuse,
            input_norm=input_norm,
            output_norm=output_norm,
            output_low=output_low,
            output_high=output_high,
            mlp_config=mlp_config,
            name_scope=name_scope
        )

        mlp_net = MLP(input_ph=state_input,
                      **mlp_kwargs,
                      net_name='deterministic_mlp_policy')

        PlaceholderInput.__init__(self, parameters=None)
        self.parameters = ParametersWithTensorflowVariable(tf_var_list=mlp_net.var_list,
                                                           rest_parameters=mlp_kwargs,
                                                           name='deterministic_mlp_policy_tf_param')
        self.state_input = state_input
        self.mlp_net = mlp_net
        self.action_tensor = mlp_net.output
        self.mlp_config = mlp_config
        self.mlp_config = mlp_config
        self.input_norm = input_norm
        self.output_norm = output_norm
        self.output_low = output_low
        self.output_high = output_high
        self.name_scope = name_scope

    @overrides.overrides
    def forward(self, obs: (np.ndarray, list), sess=None, feed_dict=None, **kwargs):
        obs = make_batch(obs, original_shape=self.env_spec.obs_shape)
        feed_dict = {} if feed_dict is None else feed_dict
        feed_dict = {
            **feed_dict,
            self.state_input: obs,
            **self.parameters.return_tf_parameter_feed_dict()
        }
        sess = sess if sess else tf.get_default_session()
        res = sess.run(self.action_tensor, feed_dict=feed_dict)
        res = np.clip(res, a_min=self.env_spec.action_space.low, a_max=self.env_spec.action_space.high)
        return res

    @overrides.overrides
    def copy_from(self, obj) -> bool:
        return PlaceholderInput.copy_from(self, obj)

    def make_copy(self, *args, **kwargs):
        kwargs = _get_copy_arg_with_tf_reuse(obj=self, kwargs=kwargs)

        copy_mlp_policy = DeterministicMLPPolicy(env_spec=self.env_spec,
                                                 input_norm=self.input_norm,
                                                 output_norm=self.output_norm,
                                                 output_low=self.output_low,
                                                 output_high=self.output_high,
                                                 mlp_config=self.mlp_config,
                                                 **kwargs)
        return copy_mlp_policy

    def save(self, *args, **kwargs):
        return PlaceholderInput.save(self, *args, **kwargs)

    def load(self, *args, **kwargs):
        return PlaceholderInput.load(self, *args, **kwargs)



from baconian.benchmark.ppo_benchmark.halfCheetah import half_cheetah_task_fn
from baconian.benchmark.ppo_benchmark.pendulum import pendulum_task_fn
from baconian.benchmark.ppo_benchmark.reacher import reacher_task_fn
from baconian.benchmark.ppo_benchmark.swimmer import swimmer_task_fn
from baconian.benchmark.ppo_benchmark.hopper import hopper_task_fn
from baconian.benchmark.ppo_benchmark.inverted_pendulum import inverted_pendulum_task_fn
from baconian.benchmark.ppo_benchmark.halfCheetah_pybullet import \
    half_cheetah_task_fn as half_cheetah_bullet_env_task_fn



from baconian.common.spaces.base import Space
from baconian.common.spaces.box import Box
from baconian.common.spaces.dict import Dict
from baconian.common.spaces.discrete import Discrete
from baconian.common.spaces.tuple import Tuple

__all__ = ["Space", "Box", "Dict", "Discrete", "Tuple"]



"""
In this example, we demonstrate how to utilize the scheduler module to dynamically setting the
learning rate of your algorithm, or epsilon-greedy probability
"""

from baconian.algo.dqn import DQN
from baconian.core.core import EnvSpec
from baconian.envs.gym_env import make
from baconian.algo.value_func.mlp_q_value import MLPQValueFunction
from baconian.core.agent import Agent
from baconian.algo.misc import EpsilonGreedy
from baconian.core.experiment import Experiment
from baconian.core.flow.train_test_flow import create_train_test_flow
from baconian.config.global_config import GlobalConfig
from baconian.common.schedules import LinearScheduler, PiecewiseScheduler, PeriodicalEventSchedule
from baconian.core.status import get_global_status_collect


def task_fn():
    env = make('Acrobot-v1')
    name = 'example_scheduler_'
    env_spec = EnvSpec(obs_space=env.observation_space,
                       action_space=env.action_space)

    mlp_q = MLPQValueFunction(env_spec=env_spec,
                              name_scope=name + '_mlp_q',
                              name=name + '_mlp_q',
                              mlp_config=[
                                  {
                                      "ACT": "RELU",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "1",
                                      "N_UNITS": 16,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  },
                                  {
                                      "ACT": "LINEAR",
                                      "B_INIT_VALUE": 0.0,
                                      "NAME": "OUPTUT",
                                      "N_UNITS": 1,
                                      "TYPE": "DENSE",
                                      "W_NORMAL_STDDEV": 0.03
                                  }
                              ])
    dqn = DQN(env_spec=env_spec,
              config_or_config_dict=dict(REPLAY_BUFFER_SIZE=1000,
                                         GAMMA=0.99,
                                         BATCH_SIZE=10,
                                         LEARNING_RATE=0.001,
                                         TRAIN_ITERATION=1,
                                         DECAY=0.5),
              name=name + '_dqn',
              value_func=mlp_q)
    agent = Agent(env=env, env_spec=env_spec,
                  algo=dqn,
                  name=name + '_agent',
                  algo_saving_scheduler=PeriodicalEventSchedule(
                      t_fn=lambda: get_global_status_collect()('TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                      trigger_every_step=20,
                      after_t=10),
                  exploration_strategy=EpsilonGreedy(action_space=env_spec.action_space,
                                                     prob_scheduler=PiecewiseScheduler(
                                                         t_fn=lambda: get_global_status_collect()(
                                                             'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'),
                                                         endpoints=((10, 0.3), (100, 0.1), (200, 0.0)),
                                                         outside_value=0.0
                                                     ),
                                                     init_random_prob=0.5))
    flow = create_train_test_flow(
        test_every_sample_count=10,
        train_every_sample_count=10,
        start_test_after_sample_count=5,
        start_train_after_sample_count=5,
        train_func_and_args=(agent.train, (), dict()),
        test_func_and_args=(agent.test, (), dict(sample_count=10)),
        sample_func_and_args=(agent.sample, (), dict(sample_count=100,
                                                     env=agent.env,
                                                     store_flag=True))
    )
    experiment = Experiment(
        tuner=None,
        env=env,
        agent=agent,
        flow=flow,
        name=name + 'experiment_debug'
    )

    dqn.parameters.set_scheduler(param_key='LEARNING_RATE',
                                 scheduler=LinearScheduler(
                                     t_fn=experiment.TOTAL_AGENT_TRAIN_SAMPLE_COUNT,
                                     schedule_timesteps=GlobalConfig().DEFAULT_EXPERIMENT_END_POINT[
                                         'TOTAL_AGENT_TRAIN_SAMPLE_COUNT'],
                                     final_p=0.0001,
                                     initial_p=0.01))
    experiment.run()


from baconian.core.experiment_runner import single_exp_runner

GlobalConfig().set('DEFAULT_LOG_PATH', './log_path')
single_exp_runner(task_fn, del_if_log_path_existed=True)



import numpy as np
from baconian.common.spaces.base import Space


class Discrete(Space):
    """
    {0,1,...,n-1}
    """

    def __init__(self, n):
        self._n = n

    @property
    def n(self):
        return self._n

    def sample(self):
        return np.random.randint(self.n)

    def contains(self, x):
        if float(x).is_integer() is False:
            return False
        x = np.array(x, dtype=np.int)
        return bool(x.shape == () and x.dtype.kind == 'i' and x >= 0 and x < self.n)

    def __repr__(self):
        return "Discrete(%d)" % self.n

    def __eq__(self, other):
        if not isinstance(other, Discrete):
            return False
        return self.n == other.n

    def flatten(self, x):
        import baconian.common.special as special

        return special.to_onehot(x, self.n)

    def unflatten(self, x):
        import baconian.common.special as special

        return special.from_onehot(x)

    def flatten_n(self, x):
        import baconian.common.special as special

        return special.to_onehot_n(x, self.n)

    def unflatten_n(self, x):
        import baconian.common.special as special

        return special.from_onehot_n(x)

    @property
    def flat_dim(self):
        return self.n

    def weighted_sample(self, weights):
        import baconian.common.special as special

        return special.weighted_sample(weights, range(self.n))

    @property
    def default_value(self):
        return 0

    def __hash__(self):
        return hash(self.n)

    def new_tensor_variable(self, name, extra_dims):
        raise NotImplementedError

    def clip(self, x):
        x = np.asarray(x).astype(np.int)
        assert x.shape == ()
        if self.contains(x):
            return x
        else:
            if x < 0:
                return 0
            elif x >= self.n:
                return self.n - 1

    def bound(self):
        return 0, self.n



from unittest import TestLoader, TextTestRunner, TestSuite
import sys
import os

path = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(path)
# print('join {} into environ path'.format(path))
src_dir = os.path.abspath(os.path.join(path, os.pardir, os.pardir))
sys.path.append(src_dir)
print('join {} into environ path'.format(src_dir))


def test_all(dir=''):
    loader = TestLoader()
    suite = TestSuite()
    for all_test_suite in loader.discover(start_dir=os.path.join(path, 'tests', dir), pattern='test*.py'):
        for test_case in all_test_suite:
            suite.addTest(test_case)
    TextTestRunner().run(test=suite)


if __name__ == '__main__':
    test_all()



from baconian.common.plotter import Plotter
import pandas as pd
from baconian.common.files import *
from collections import OrderedDict
from typing import Union


class SingleExpLogDataLoader(object):
    def __init__(self, exp_root_dir: str):
        self._root_dir = exp_root_dir
        check_file(path=os.path.join(exp_root_dir, 'record', 'final_status.json'))  # dir list
        check_file(path=os.path.join(exp_root_dir, 'record', 'global_config.json'))
        self.final_status = load_json(file_path=os.path.join(exp_root_dir, 'record', 'final_status.json'))
        self.global_config = load_json(file_path=os.path.join(exp_root_dir, 'record', 'global_config.json'))

    def load_record_data(self, agent_log_dir_name, algo_log_dir_name, env_log_dir_name):
        # TODO  pre-load all data here
        check_dir(os.path.join(self._root_dir, agent_log_dir_name))
        check_dir(os.path.join(self._root_dir, algo_log_dir_name))

    def init(self):
        pass

    def plot_res(self, sub_log_dir_name, key, index, save_path=None, mode=('line', 'hist', 'scatter'),
                 average_over=1, file_name=None, save_format='png', save_flag=False,
                 ):
        log_name = os.path.join(self._root_dir, 'record', sub_log_dir_name, 'log.json')
        f = open(log_name, 'r')
        res_dict = json.load(f)
        key_list = res_dict[key]
        key_value = OrderedDict()
        key_vector = []
        index_vector = []
        for record in key_list:
            num_index = int(record[index])
            index_vector.append(num_index)
            key_vector.append(record["log_val"])
        key_value[index] = index_vector
        key_value[key] = key_vector
        data = pd.DataFrame.from_dict(key_value)  # Create dataframe for plotting
        row_num = data.shape[0]
        column_num = data.shape[1]
        data_new = data

        # Calculate mean value in horizontal axis, incompatible with histogram mode
        if average_over != 1:
            if mode != 'histogram':
                new_row_num = int(row_num / average_over)
                data_new = data.head(new_row_num).copy()
                data_new.loc[:, index] = data_new.loc[:, index] * average_over
                for column in range(1, column_num):
                    for i in range(new_row_num):
                        data_new.iloc[i, column] = data.iloc[i * average_over: i * average_over + average_over,
                                                   column].mean()

        if mode == 'histogram':
            histogram_flag = True
            data_new = data.iloc[:, 1:].copy()
        else:
            histogram_flag = False
        scatter_flag = True if mode == 'scatter' else False

        Plotter.plot_any_key_in_log(data=data_new, index=index, key=key,
                                    sub_log_dir_name=sub_log_dir_name,
                                    scatter_flag=scatter_flag, save_flag=save_flag,
                                    histogram_flag=histogram_flag, save_path=save_path,
                                    save_format=save_format, file_name=file_name)


# TODO
# key and index, log given = plot figure(curve), average or 10, or set a range
# normalisation

class MultipleExpLogDataLoader(object):
    def __init__(self, exp_root_dir_list: Union[str, list]):
        self._root_dir = exp_root_dir_list
        self.exp_list = []
        self.num = 0
        if type(self._root_dir) is str:
            for path in os.listdir(self._root_dir):
                print('path: ', path)
                exp_root_dir = os.path.join(self._root_dir, path)
                print(exp_root_dir)
                self.exp_list.append(exp_root_dir)
                self.num += 1
                SingleExpLogDataLoader(exp_root_dir)
        else:
            for dependent_exp in self._root_dir:
                assert type(dependent_exp) is str
                for path in os.listdir(dependent_exp):
                    exp_root_dir = os.path.join(dependent_exp, path)
                    self.exp_list.append(exp_root_dir)
                    self.num += 1
                    SingleExpLogDataLoader(exp_root_dir)

    def plot_res(self, key, index, save_path, sub_log_dir_name: str, mode=('plot', 'hist', 'scatter'), average_over=1,
                 save_format='png', file_name=None, save_flag=False):
        multiple_key_value = {}
        for exp in self.exp_list:
            f = open(os.path.join(exp, 'record', sub_log_dir_name, 'log.json'), 'r')
            res_dict = json.load(f)
            key_list = res_dict[key]
            key_vector = []
            index_vector = []
            for record in key_list:
                num_index = int(record[index])
                index_vector.append(num_index)
                key_vector.append(record["log_val"])
            multiple_key_value[index] = index_vector
            multiple_key_value[key + '_' + exp] = key_vector
        data = pd.DataFrame.from_dict(multiple_key_value)  # Create dataframe for plotting
        row_num = data.shape[0]
        column_num = data.shape[1]
        data_new = data

        # Calculate mean value in horizontal axis, incompatible with histogram mode
        if average_over != 1:
            if mode != 'histogram':
                new_row_num = int(row_num / average_over)
                data_new = data.head(new_row_num).copy()
                data_new.loc[:, index] = data_new.loc[:, index] * average_over
                for column in range(1, column_num):
                    for i in range(new_row_num):
                        data_new.iloc[i, column] = data.iloc[i * average_over: i * average_over + average_over,
                                                   column].mean()

        data_new['MEAN'] = data_new[data_new.columns[1:]].mean(axis=1)  # axis = 1 in columns, first column not counted
        data_new['STD_DEV'] = data_new[data_new.columns[1:-1]].std(axis=1)

        if mode == 'histogram':
            histogram_flag = True
            data_new = data.iloc[:, 1:-2].copy().stack()  # Mean and variance columns not counted
        else:
            histogram_flag = False
        if mode == 'scatter':
            scatter_flag = True
        else:
            scatter_flag = False

        Plotter.plot_any_key_in_log(data=data_new, index=index, key=key, exp_num=self.num,
                                    scatter_flag=scatter_flag, save_flag=save_flag,
                                    mean_stddev_flag=True,
                                    histogram_flag=histogram_flag, save_path=save_path,
                                    sub_log_dir_name=sub_log_dir_name, save_format=save_format, file_name=file_name)



import os
import shutil

import numpy as np
import tensorflow as tf

from baconian.common.logging import Logger, ConsoleLogger
from baconian.config.global_config import GlobalConfig
from baconian.core.global_var import reset_all, get_all
from baconian.tf.util import create_new_tf_session
from baconian.test.tests.set_up.class_creator import ClassCreatorSetup
from baconian.core.status import reset_global_status_collect


class BaseTestCase(ClassCreatorSetup):
    def setUp(self):
        reset_all()
        reset_global_status_collect()
        for key, val in get_all().items():
            self.assertTrue(len(val) == 0)

    def tearDown(self):
        reset_all()


class TestTensorflowSetup(BaseTestCase):
    default_id = 0

    def setUp(self):
        BaseTestCase.setUp(self)
        if tf.get_default_session():
            sess = tf.get_default_session()
            sess.__exit__(None, None, None)
        tf.reset_default_graph()
        print('set tf device as {}'.format(self.default_id))
        os.environ["CUDA_VISIBLE_DEVICES"] = str(self.default_id)
        self.sess = create_new_tf_session()

    def tearDown(self):
        if self.sess.run(tf.report_uninitialized_variables()).shape[0] != 0:
            print('some variables are not uninitialized:')
            print(self.sess.run(tf.report_uninitialized_variables()))
            print(self.sess.run(tf.report_uninitialized_variables()).shape)
            raise AssertionError('some variables are not uninitialized')

        if tf.get_default_session():
            sess = tf.get_default_session()
            sess.close()
        BaseTestCase.tearDown(self)

    def assert_var_list_equal(self, var_list1, var_list2):
        for var1, var2 in zip(var_list1, var_list2):
            res1, res2 = self.sess.run([var1, var2])
            self.assertTrue(np.equal(res1, res2).all())

    def assert_var_list_at_least_not_equal(self, var_list1, var_list2):
        res = []
        for var1, var2 in zip(var_list1, var_list2):
            res1, res2 = self.sess.run([var1, var2])
            res.append(np.equal(res1, res2).all())
        self.assertFalse(np.array(res).all())

    def assert_var_list_id_equal(self, var_list1, var_list2):
        for var1, var2 in zip(var_list1, var_list2):
            self.assertTrue(id(var1) == id(var2))

    def assert_var_list_id_no_equal(self, var_list1, var_list2):
        for var1, var2 in zip(var_list1, var_list2):
            self.assertTrue(id(var1) != id(var2))


class TestWithLogSet(BaseTestCase):
    def setUp(self):
        BaseTestCase.setUp(self)
        try:
            shutil.rmtree(GlobalConfig().DEFAULT_LOG_PATH)
        except FileNotFoundError:
            pass
        # os.makedirs(GlobalConfig().DEFAULT_LOG_PATH)
        # self.assertFalse(ConsoleLogger().inited_flag)
        # self.assertFalse(Logger().inited_flag)

        Logger().init(config_or_config_dict=GlobalConfig().DEFAULT_LOG_CONFIG_DICT,
                      log_path=GlobalConfig().DEFAULT_LOG_PATH,
                      log_level=GlobalConfig().DEFAULT_LOG_LEVEL)
        ConsoleLogger().init(logger_name='console_logger',
                             to_file_flag=True,
                             level=GlobalConfig().DEFAULT_LOG_LEVEL,
                             to_file_name=os.path.join(Logger().log_dir, 'console.log'))

        self.assertTrue(ConsoleLogger().inited_flag)
        self.assertTrue(Logger().inited_flag)

    def tearDown(self):
        Logger().reset()
        ConsoleLogger().reset()
        BaseTestCase.tearDown(self)
        self.assertFalse(ConsoleLogger().inited_flag)
        self.assertFalse(Logger().inited_flag)


class TestWithAll(TestTensorflowSetup, TestWithLogSet):
    def setUp(self):
        TestTensorflowSetup.setUp(self)
        TestWithLogSet.setUp(self)

    def tearDown(self):
        TestWithLogSet.tearDown(self)
        TestTensorflowSetup.tearDown(self)


class SimpleTestSetup(BaseTestCase):
    def setUp(self):
        BaseTestCase.setUp(self)
        try:
            shutil.rmtree(GlobalConfig().DEFAULT_LOG_PATH)
        except FileNotFoundError:
            pass
        os.makedirs(GlobalConfig().DEFAULT_LOG_PATH)
        self.assertFalse(ConsoleLogger().inited_flag)
        self.assertFalse(Logger().inited_flag)

    def tearDown(self):
        BaseTestCase.tearDown(self)



import glob
import os
import importlib.util

ABS_PATH = os.path.dirname(os.path.realpath(__file__))


def test_all():
    file_list = glob.glob(os.path.join(ABS_PATH, '*.py'))
    file_list.remove(os.path.realpath(__file__))
    for f in file_list:
        spec = importlib.util.spec_from_file_location('', f)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)


if __name__ == '__main__':
    test_all()



from baconian.algo.policy.policy import DeterministicPolicy
from baconian.core.core import EnvSpec
from baconian.core.parameters import Parameters
from baconian.config.global_config import GlobalConfig
from baconian.config.dict_config import DictConfig
from baconian.common.misc import *
from copy import deepcopy


class ConstantActionPolicy(DeterministicPolicy):
    required_key_dict = DictConfig.load_json(file_path=GlobalConfig().DEFAULT_CONSTANT_ACTION_POLICY_REQUIRED_KEY_LIST)

    def __init__(self, env_spec: EnvSpec, config_or_config_dict: (DictConfig, dict), name='policy'):
        config = construct_dict_config(config_or_config_dict, self)
        parameters = Parameters(parameters=dict(),
                                source_config=config)
        assert env_spec.action_space.contains(x=config('ACTION_VALUE'))
        super().__init__(env_spec, parameters, name)
        self.config = config

    def forward(self, *args, **kwargs):
        action = self.env_spec.action_space.unflatten(self.parameters('ACTION_VALUE'))
        assert self.env_spec.action_space.contains(x=action)
        return action

    def copy_from(self, obj) -> bool:
        super().copy_from(obj)
        self.parameters.copy_from(obj.parameters)
        return True

    def make_copy(self, *args, **kwargs):
        return ConstantActionPolicy(env_spec=self.env_spec,
                                    config_or_config_dict=deepcopy(self.config),
                                    *args, **kwargs)
